import pytest

from tests._dummies.dataset import DummyClassificationDataset
from torch_uncertainty.datamodules import TinyImageNetDataModule
from torch_uncertainty.datasets.classification import TinyImageNet


class TestTinyImageNetDataModule:
    """Testing the TinyImageNetDataModule datamodule class."""

    def test_tiny_imagenet(self):
        dm = TinyImageNetDataModule(root="./data/", batch_size=128)

        assert dm.dataset == TinyImageNet

        dm = TinyImageNetDataModule(
            root="./data/",
            batch_size=128,
            rand_augment_opt="rand-m9-n3-mstd0.5",
            ood_ds="imagenet-o",
        )

        dm = TinyImageNetDataModule(
            root="./data/",
            batch_size=128,
            ood_ds="textures",
            basic_augment=False,
        )

        with pytest.raises(ValueError):
            TinyImageNetDataModule(root="./data/", batch_size=128, ood_ds="other")

        dm.dataset = DummyClassificationDataset
        dm.ood_dataset = DummyClassificationDataset
        dm.shift_dataset = DummyClassificationDataset

        dm.prepare_data()
        dm.setup()

        dm.train_dataloader()
        dm.val_dataloader()
        dm.test_dataloader()

        with pytest.raises(ValueError):
            dm.setup("other")

        dm.eval_ood = True
        dm.eval_shift = True
        dm.prepare_data()
        dm.setup("test")
        dm.test_dataloader()

        dm = TinyImageNetDataModule(root="./data/", batch_size=128, ood_ds="svhn")
        dm.dataset = DummyClassificationDataset
        dm.ood_dataset = DummyClassificationDataset
        dm.shift_dataset = DummyClassificationDataset
        dm.eval_ood = True
        dm.eval_shift = True
        dm.prepare_data()
        dm.setup("test")

    def test_tiny_imagenet_cv(self):
        dm = TinyImageNetDataModule(root="./data/", batch_size=128)
        dm.dataset = lambda root, split, transform: DummyClassificationDataset(
            root, split=split, transform=transform, num_images=20
        )
        dm.make_cross_val_splits(2, 1)

        dm = TinyImageNetDataModule(root="./data/", batch_size=128, val_split=0.1)
        dm.dataset = lambda root, split, transform: DummyClassificationDataset(
            root, split=split, transform=transform, num_images=20
        )
        dm.make_cross_val_splits(2, 1)



import logging
from collections.abc import Callable
from pathlib import Path
from typing import Any, Literal

import numpy as np
from torchvision.datasets import VisionDataset
from torchvision.datasets.utils import (
    check_integrity,
    download_and_extract_archive,
)


class MNISTC(VisionDataset):
    """The corrupted MNIST-C Dataset.

    Args:
        root (str): Root directory of the datasets.
        transform (callable, optional): A function/transform that takes in
            a PIL image and returns a transformed version. E.g,
            ``transforms.RandomCrop``. Defaults to None.
        target_transform (callable, optional): A function/transform that
            takes in the target and transforms it. Defaults to None.
        subset (str): The subset to use, one of ``all`` or the keys in
            ``mnistc_subsets``.
        download (bool, optional): If True, downloads the dataset from the
            internet and puts it in root directory. If dataset is already
            downloaded, it is not downloaded again. Defaults to False.

    References:
        Mu, Norman, and Justin Gilmer. "MNIST-C: A robustness benchmark for
        computer vision." In ICMLW 2019.

    License:
        The dataset is released by the dataset's authors under the Creative
        Commons Attribution 4.0.

    Note:
        This dataset does not contain severity levels. Raise an issue if you
        want someone to investigate this.
    """

    base_folder = "mnist_c"
    zip_md5 = "4b34b33045869ee6d424616cd3a65da3"
    mnistc_subsets = [
        "brightness",
        "canny_edges",
        "dotted_line",
        "fog",
        "glass_blur",
        "impulse_noise",
        "motion_blur",
        "rotate",
        "scale",
        "shear",
        "shot_noise",
        "spatter",
        "stripe",
        "translate",
        "zigzag",
    ]

    url = "https://zenodo.org/record/3239543/files/mnist_c.zip"
    filename = "mnist_c.zip"

    def __init__(
        self,
        root: str | Path,
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        split: Literal["train", "test"] = "test",
        subset: str = "all",
        download: bool = False,
    ) -> None:
        self.root = Path(root)

        # Download the new targets
        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found. You can use download=True to download it.")

        super().__init__(
            root=self.root / self.base_folder,
            transform=transform,
            target_transform=target_transform,
        )
        if subset not in ["all", *self.mnistc_subsets]:
            raise ValueError(f"The subset '{subset}' does not exist in MNIST-C.")
        self.subset = subset

        if split not in ["train", "test"]:
            raise ValueError(f"The split '{split}' should be either 'train' or 'test'.")
        self.split = split

        samples, labels = self.make_dataset(self.root, self.subset, self.split)

        self.samples = samples
        self.labels = labels

    def make_dataset(
        self,
        root: Path,
        subset: str,
        split: Literal["train", "test"],
    ) -> tuple[np.ndarray, np.ndarray]:
        r"""Build the corrupted dataset according to the chosen subset and
            severity. If the subset is 'all', gather all corruption types
            in the dataset.

        Args:
            root (Path):The path to the dataset.
            subset (str): The name of the corruption subset to be used. Choose
                `all` for the dataset to contain all subsets.
            split (str): The split to be used, either `train` or `test`.

        Returns:
            Tuple[np.ndarray, np.ndarray]: The samples and labels of the chosen.
        """
        if subset == "all":
            # take any subset to get the labels
            labels: np.ndarray = np.load(root / f"identity/{split}_labels.npy")

            sample_arrays = [
                np.load(root / mnist_subset / f"{split}_images.npy")
                for mnist_subset in self.mnistc_subsets
            ]
            samples = np.concatenate(sample_arrays, axis=0)
            labels = np.tile(labels, len(self.mnistc_subsets))

        else:
            samples: np.ndarray = np.load(root / subset / f"{split}_images.npy")
            labels: np.ndarray = np.load(root / f"{split}_labels.npy")
        return samples, labels

    def __len__(self) -> int:
        """The number of samples in the dataset."""
        return self.labels.shape[0]

    def __getitem__(self, index: int) -> Any:
        """Get the samples and targets of the dataset.

        Args:
            index (int): The index of the sample to get.
        """
        sample, target = (
            self.samples[index],
            self.labels[index],
        )

        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return sample, target

    def _check_integrity(self) -> bool:
        """Check the integrity of the dataset."""
        fpath = self.root / self.filename
        return check_integrity(fpath, self.zip_md5)

    def download(self) -> None:
        """Download the dataset."""
        if self._check_integrity():
            logging.info("Files already downloaded and verified")
            return
        download_and_extract_archive(self.url, self.root, filename=self.filename, md5=self.zip_md5)



import inspect

import torch
import torch.nn.functional as F
from torch import Tensor, nn


def get_dist_linear_layer(dist_family: str) -> type[nn.Module]:
    if dist_family == "normal":
        return NormalLinear
    if dist_family == "laplace":
        return LaplaceLinear
    if dist_family == "cauchy":
        return CauchyLinear
    if dist_family == "student":
        return StudentTLinear
    if dist_family == "nig":
        return NormalInverseGammaLinear
    raise NotImplementedError(
        f"{dist_family} distribution is not supported. Raise an issue if needed."
    )


def get_dist_conv_layer(dist_family: str) -> type[nn.Module]:
    if dist_family == "normal":
        return NormalConvNd
    if dist_family == "laplace":
        return LaplaceConvNd
    if dist_family == "cauchy":
        return CauchyConvNd
    if dist_family == "student":
        return StudentTConvNd
    if dist_family == "nig":
        return NormalInverseGammaConvNd
    raise NotImplementedError(
        f"{dist_family} distribution is not supported. Raise an issue if needed."
    )


class _ExpandOutputLinear(nn.Module):
    """Abstract class for expanding the output of any nn.Module using an `out_features` argument.

    Args:
        base_layer (type[nn.Module]): The base layer class.
        event_dim (int): The number of event dimensions.
        num_params (int): The number of parameters to output. For instance, the normal distribution
            has 2 parameters (loc and scale).
        **layer_args: Additional arguments for the base layer.
    """

    def __init__(self, base_layer: type[nn.Module], event_dim: int, num_params: int, **layer_args):
        if "out_features" not in inspect.getfullargspec(base_layer.__init__).args:
            raise ValueError(f"{base_layer.__name__} does not have an `out_features` argument.")

        super().__init__()
        self.base_layer = base_layer(out_features=num_params * event_dim, **layer_args)
        self.event_dim = event_dim

    def forward(self, x: Tensor) -> Tensor:
        return self.base_layer(x)


class _ExpandOutputConvNd(nn.Module):
    """Abstract class for expanding the output of any nn.Module using an `out_channels` argument.

    Args:
        base_layer (type[nn.Module]): The base layer class.
        event_dim (int): The number of event dimensions.
        num_params (int): The number of parameters to output. For instance, the normal distribution
            has 2 parameters (loc and scale).
        **layer_args: Additional arguments for the base layer.
    """

    def __init__(self, base_layer: type[nn.Module], event_dim: int, num_params: int, **layer_args):
        if "out_channels" not in inspect.getfullargspec(base_layer.__init__).args:
            raise ValueError(f"{base_layer.__name__} does not have an `out_channels` argument.")

        super().__init__()
        self.base_layer = base_layer(out_channels=num_params * event_dim, **layer_args)
        self.event_dim = event_dim

    def forward(self, x: Tensor) -> Tensor:
        return self.base_layer(x)


class _LocScaleLinear(_ExpandOutputLinear):
    """Base Linear layer for any distribution with loc and scale parameters.

    Args:
        base_layer (type[nn.Module]): The base layer class.
        event_dim (int): The number of event dimensions.
        min_scale (float): The minimal value of the scale parameter.
        **layer_args: Additional arguments for the base layer.
    """

    def __init__(
        self,
        base_layer: type[nn.Module],
        event_dim: int,
        min_scale: float = 1e-6,
        **layer_args,
    ) -> None:
        super().__init__(
            base_layer=base_layer,
            event_dim=event_dim,
            num_params=2,
            **layer_args,
        )
        self.min_scale = min_scale

    def forward(self, x: Tensor) -> dict[str, Tensor]:
        x = super().forward(x)
        loc = x[..., : self.event_dim]
        scale = torch.clamp(
            F.softplus(x[..., self.event_dim : 2 * self.event_dim]), min=self.min_scale
        )
        return {"loc": loc, "scale": scale}


class _LocScaleConvNd(_ExpandOutputConvNd):
    """Base Convolutional layer for any distribution with loc and scale parameters.

    Args:
        base_layer (type[nn.Module]): The base layer class.
        event_dim (int): The number of event dimensions.
        min_scale (float): The minimal value of the scale parameter.
        **layer_args: Additional arguments for the base layer.
    """

    def __init__(
        self,
        base_layer: type[nn.Module],
        event_dim: int,
        min_scale: float = 1e-6,
        **layer_args,
    ) -> None:
        super().__init__(
            base_layer=base_layer,
            event_dim=event_dim,
            num_params=2,
            **layer_args,
        )
        self.min_scale = min_scale

    def forward(self, x: Tensor) -> dict[str, Tensor]:
        x = super().forward(x)
        loc = x[:, : self.event_dim]
        scale = torch.clamp(
            F.softplus(x[:, self.event_dim : 2 * self.event_dim]), min=self.min_scale
        )
        return {"loc": loc, "scale": scale}


class NormalLinear(_LocScaleLinear):
    r"""Normal Distribution Linear Density Layer.

    Args:
        base_layer (type[nn.Module]): The base layer class.
        event_dim (int): The number of event dimensions.
        min_scale (float): The minimal value of the scale parameter.
        **layer_args: Additional arguments for the base layer.

    Shape:
        - Input: :math:`(\ast, H_{in})` where :math:`\ast` means any number of dimensions including
          none and :math:`H_{in} = \text{in_features}`.
        - Output: A dict with the following keys

          - ``"loc"``: The mean of the Normal distribution of shape :math:`(\ast, H_{out})` where
            all but the last dimension are the same as the input and
            :math:`H_{out} = \text{out_features}`.
          - ``"scale"``: The standard deviation of the Normal distribution of shape
            :math:`(\ast, H_{out})`.
    """


class NormalConvNd(_LocScaleConvNd):
    r"""Normal Distribution Convolutional Density Layer.

    Args:
        in_channels (int): The number of input channels.
        out_channels (int): The number of event channels.
        kernel_size (int | tuple[int]): The size of the convolutional kernel.
        stride (int | tuple[int]): The stride of the convolution.
        padding (int | tuple[int]): The padding of the convolution.
        dilation (int | tuple[int]): The dilation of the convolution.
        groups (int): The number of groups in the convolution.
        min_scale (float): The minimal value of the scale parameter.
        device (torch.device): The device where the layer is stored.
        dtype (torch.dtype): The datatype of the layer.

    Shape:
        - Input: :math:`(N, C_{in}, \ast)` where :math:`\ast` means any number of dimensions and
          :math:`C_{in} = \text{in_channels}` and :math:`N` is the batch size.
        - Output: A dict with the following keys

          - ``"loc"``: The mean of the Normal distribution of shape :math:`(N, C_{out}, \ast)` where
            :math:`C_{out} = \text{out_channels}`.
          - ``"scale"``: The standard deviation of the Normal distribution of shape
            :math:`(\ast, C_{out}, \ast)`.
    """


class LaplaceLinear(_LocScaleLinear):
    r"""Laplace Distribution Linear Density Layer.

    Args:
        base_layer (type[nn.Module]): The base layer class.
        event_dim (int): The number of event dimensions.
        min_scale (float): The minimal value of the scale parameter.
        **layer_args: Additional arguments for the base layer.

    Shape:
        - Input: :math:`(\ast, H_{in})` where :math:`\ast` means any number of dimensions including
          none and :math:`H_{in} = \text{in_features}`.
        - Output: A dict with the following keys

          - ``"loc"``: The mean of the Laplace distribution of shape :math:`(\ast, H_{out})` where
            all but the last dimension are the same as the input and
            :math:`H_{out} = \text{out_features}`.
          - ``"scale"``: The standard deviation of the Laplace distribution of shape
            :math:`(\ast, H_{out})`.
    """


class LaplaceConvNd(_LocScaleConvNd):
    r"""Laplace Distribution Convolutional Density Layer.

    Args:
        in_channels (int): The number of input channels.
        out_channels (int): The number of event channels.
        kernel_size (int | tuple[int]): The size of the convolutional kernel.
        stride (int | tuple[int]): The stride of the convolution.
        padding (int | tuple[int]): The padding of the convolution.
        dilation (int | tuple[int]): The dilation of the convolution.
        groups (int): The number of groups in the convolution.
        min_scale (float): The minimal value of the scale parameter.
        device (torch.device): The device where the layer is stored.
        dtype (torch.dtype): The datatype of the layer.

    Shape:
        - Input: :math:`(N, C_{in}, \ast)` where :math:`\ast` means any number of dimensions and
          :math:`C_{in} = \text{in_channels}` and :math:`N` is the batch size.
        - Output: A dict with the following keys

          - ``"loc"``: The mean of the Laplace distribution of shape :math:`(N, C_{out}, \ast)` where
            :math:`C_{out} = \text{out_channels}`.
          - ``"scale"``: The standard deviation of the Laplace distribution of shape
            :math:`(\ast, C_{out}, \ast)`.
    """


class CauchyLinear(_LocScaleLinear):
    r"""Cauchy Distribution Linear Density Layer.

    Args:
        base_layer (type[nn.Module]): The base layer class.
        event_dim (int): The number of event dimensions.
        min_scale (float): The minimal value of the scale parameter.
        **layer_args: Additional arguments for the base layer.

    Shape:
        - Input: :math:`(\ast, H_{in})` where :math:`\ast` means any number of dimensions including
          none and :math:`H_{in} = \text{in_features}`.
        - Output: A dict with the following keys

          - ``"loc"``: The mean of the Cauchy distribution of shape :math:`(\ast, H_{out})` where
            all but the last dimension are the same as the input and
            :math:`H_{out} = \text{out_features}`.
          - ``"scale"``: The standard deviation of the Cauchy distribution of shape
            :math:`(\ast, H_{out})`.
    """


class CauchyConvNd(_LocScaleConvNd):
    r"""Cauchy Distribution Convolutional Density Layer.

    Args:
        in_channels (int): The number of input channels.
        out_channels (int): The number of event channels.
        kernel_size (int | tuple[int]): The size of the convolutional kernel.
        stride (int | tuple[int]): The stride of the convolution.
        padding (int | tuple[int]): The padding of the convolution.
        dilation (int | tuple[int]): The dilation of the convolution.
        groups (int): The number of groups in the convolution.
        min_scale (float): The minimal value of the scale parameter.
        device (torch.device): The device where the layer is stored.
        dtype (torch.dtype): The datatype of the layer.

    Shape:
        - Input: :math:`(N, C_{in}, \ast)` where :math:`\ast` means any number of dimensions and
          :math:`C_{in} = \text{in_channels}` and :math:`N` is the batch size.
        - Output: A dict with the following keys

          - ``"loc"``: The mean of the Cauchy distribution of shape :math:`(N, C_{out}, \ast)` where
            :math:`C_{out} = \text{out_channels}`.
          - ``"scale"``: The standard deviation of the Cauchy distribution of shape
            :math:`(\ast, C_{out}, \ast)`.
    """


class StudentTLinear(_ExpandOutputLinear):
    r"""Student's T-Distribution Linear Density Layer.

    Args:
        base_layer (type[nn.Module]): The base layer class.
        event_dim (int): The number of event dimensions.
        min_scale (float): The minimal value of the scale parameter.
        min_df (float): The minimal value of the degrees of freedom parameter.
        fixed_df (float): If not None, the degrees of freedom parameter is fixed to this value.
            Otherwise, it is learned.
        **layer_args: Additional arguments for the base layer.

    Shape:
        - Input: :math:`(\ast, H_{in})` where :math:`\ast` means any number of dimensions including
          none and :math:`H_{in} = \text{in_features}`.
        - Output: A dict with the following keys

          - ``"loc"``: The mean of the Student's t-distribution of shape :math:`(\ast, H_{out})` where
            all but the last dimension are the same as the input and
            :math:`H_{out} = \text{out_features}`.
          - ``"scale"``: The standard deviation of the Student's t-distribution of shape
            :math:`(\ast, H_{out})`.
          - ``"df"``: The degrees of freedom of the Student's t distribution of shape
            :math:`(\ast, H_{out})` or Number.
    """

    def __init__(
        self,
        base_layer: type[nn.Module],
        event_dim: int,
        min_scale: float = 1e-6,
        min_df: float = 2.0,
        fixed_df: float | None = None,
        **layer_args,
    ) -> None:
        super().__init__(
            base_layer=base_layer,
            event_dim=event_dim,
            num_params=3 if fixed_df is None else 2,
            **layer_args,
        )

        self.min_scale = min_scale
        self.min_df = min_df
        self.fixed_df = fixed_df

    def forward(self, x: Tensor) -> dict[str, Tensor]:
        x = super().forward(x)
        loc = x[..., : self.event_dim]
        scale = torch.clamp(
            F.softplus(x[..., self.event_dim : 2 * self.event_dim]), min=self.min_scale
        )
        df = (
            torch.clamp(F.softplus(x[..., 2 * self.event_dim :]), min=self.min_df)
            if self.fixed_df is None
            else torch.full_like(loc, self.fixed_df)
        )
        return {"loc": loc, "scale": scale, "df": df}


class StudentTConvNd(_ExpandOutputConvNd):
    r"""Student's T-Distribution Convolutional Density Layer.

    Args:
        base_layer (type[nn.Module]): The base layer class.
        event_dim (int): The number of event dimensions.
        min_scale (float): The minimal value of the scale parameter.
        min_df (float): The minimal value of the degrees of freedom parameter.
        fixed_df (float): If not None, the degrees of freedom parameter is fixed to this value.
            Otherwise, it is learned.
        **layer_args: Additional arguments for the base layer.

    Shape:
        - Input: :math:`(N, C_{in}, \ast)` where :math:`\ast` means any number of dimensions and
          :math:`C_{in} = \text{in_channels}` and :math:`N` is the batch size.
        - Output: A dict with the following keys

          - ``"loc"``: The mean of the Student's t-distribution of shape :math:`(N, C_{out}, \ast)` where
            :math:`C_{out} = \text{out_channels}`.
          - ``"scale"``: The standard deviation of the Student's t-distribution of shape
            :math:`(\ast, C_{out}, \ast)`.
          - ``"df"``: The degrees of freedom of the Student's t distribution of shape
            :math:`(\ast, C_{out}, \ast)`.
    """

    def __init__(
        self,
        base_layer: type[nn.Module],
        event_dim: int,
        min_scale: float = 1e-6,
        min_df: float = 2.0,
        fixed_df: float | None = None,
        **layer_args,
    ) -> None:
        super().__init__(
            base_layer=base_layer,
            event_dim=event_dim,
            num_params=3 if fixed_df is None else 2,
            **layer_args,
        )

        self.min_scale = min_scale
        self.min_df = min_df
        self.fixed_df = fixed_df

    def forward(self, x: Tensor) -> dict[str, Tensor]:
        x = super().forward(x)
        loc = x[:, : self.event_dim]
        scale = torch.clamp(
            F.softplus(x[:, self.event_dim : 2 * self.event_dim]), min=self.min_scale
        )
        df = (
            torch.clamp(F.softplus(x[:, 2 * self.event_dim :]), min=self.min_df)
            if self.fixed_df is None
            else torch.full_like(loc, self.fixed_df)
        )
        return {"loc": loc, "scale": scale, "df": df}


class NormalInverseGammaLinear(_ExpandOutputLinear):
    r"""Normal-Inverse-Gamma Distribution Linear Density Layer.

    Args:
        base_layer (type[nn.Module]): The base layer class.
        event_dim (int): The number of event dimensions.
        min_lmbda (float): The minimal value of the :math:`\lambda` parameter.
        min_alpha (float): The minimal value of the :math:`\alpha` parameter.
        min_beta (float): The minimal value of the :math:`\beta` parameter.
        **layer_args: Additional arguments for the base layer.

    Shape:
        - Input: :math:`(\ast, H_{in})` where :math:`\ast` means any number of dimensions including
          none and :math:`H_{in} = \text{in_features}`.
        - Output: A dict with the following keys

          - ``"loc"``: The mean of the Normal-Inverse-Gamma distribution of shape :math:`(\ast, H_{out})` where
            all but the last dimension are the same as the input and
            :math:`H_{out} = \text{out_features}`.
          - ``"lmbda"``: The lambda parameter of the Normal-Inverse-Gamma distribution of shape
            :math:`(\ast, H_{out})`.
          - ``"alpha"``: The alpha parameter of the Normal-Inverse-Gamma distribution of shape
            :math:`(\ast, H_{out})`.
          - ``"beta"``: The beta parameter of the Normal-Inverse-Gamma distribution of shape
            :math:`(\ast, H_{out})`.

    Source:
        - `Normal-Inverse-Gamma Distribution <https://en.wikipedia.org/wiki/Normal-inverse-gamma_distribution>`_
    """

    def __init__(
        self,
        base_layer: type[nn.Module],
        event_dim: int,
        min_lmbda: float = 1e-6,
        min_alpha: float = 1e-6,
        min_beta: float = 1e-6,
        **layer_args,
    ) -> None:
        super().__init__(
            base_layer=base_layer,
            event_dim=event_dim,
            num_params=4,
            **layer_args,
        )

        self.min_lmbda = min_lmbda
        self.min_alpha = min_alpha
        self.min_beta = min_beta

    def forward(self, x: Tensor) -> dict[str, Tensor]:
        x = super().forward(x)
        loc = x[..., : self.event_dim]
        lmbda = torch.clamp(
            F.softplus(x[..., self.event_dim : 2 * self.event_dim]), min=self.min_lmbda
        )
        alpha = 1 + torch.clamp(
            F.softplus(x[..., 2 * self.event_dim : 3 * self.event_dim]), min=self.min_alpha
        )
        beta = torch.clamp(F.softplus(x[..., 3 * self.event_dim :]), min=self.min_beta)
        return {
            "loc": loc,
            "lmbda": lmbda,
            "alpha": alpha,
            "beta": beta,
        }


class NormalInverseGammaConvNd(_ExpandOutputConvNd):
    r"""Normal-Inverse-Gamma Distribution Convolutional Density Layer.

    Args:
        base_layer (type[nn.Module]): The base layer class.
        event_dim (int): The number of event dimensions.
        min_lmbda (float): The minimal value of the :math:`\lambda` parameter.
        min_alpha (float): The minimal value of the :math:`\alpha` parameter.
        min_beta (float): The minimal value of the :math:`\beta` parameter.
        **layer_args: Additional arguments for the base layer.

    Shape:
        - Input: :math:`(N, C_{in}, \ast)` where :math:`\ast` means any number of dimensions and
          :math:`C_{in} = \text{in_channels}` and :math:`N` is the batch size.
        - Output: A dict with the following keys

          - ``"loc"``: The mean of the Normal-Inverse-Gamma distribution of shape :math:`(N, C_{out}, \ast)` where
            :math:`C_{out} = \text{out_channels}`.
          - ``"lmbda"``: The lambda parameter of the Normal-Inverse-Gamma distribution of shape
            :math:`(N, C_{out}, \ast)`.
          - ``"alpha"``: The alpha parameter of the Normal-Inverse-Gamma distribution of shape
            :math:`(N, C_{out}, \ast)`.
          - ``"beta"``: The beta parameter of the Normal-Inverse-Gamma distribution of shape
            :math:`(N, C_{out}, \ast)`.

    Source:
        - `Normal-Inverse-Gamma Distribution <https://en.wikipedia.org/wiki/Normal-inverse-gamma_distribution>`_
    """

    def __init__(
        self,
        base_layer: type[nn.Module],
        event_dim: int,
        min_lmbda: float = 1e-6,
        min_alpha: float = 1e-6,
        min_beta: float = 1e-6,
        **layer_args,
    ) -> None:
        super().__init__(
            base_layer=base_layer,
            event_dim=event_dim,
            num_params=4,
            **layer_args,
        )

        self.min_lmbda = min_lmbda
        self.min_alpha = min_alpha
        self.min_beta = min_beta

    def forward(self, x: Tensor) -> dict[str, Tensor]:
        x = super().forward(x)
        loc = x[:, : self.event_dim]
        lmbda = torch.clamp(
            F.softplus(x[:, self.event_dim : 2 * self.event_dim]), min=self.min_lmbda
        )
        alpha = 1 + torch.clamp(
            F.softplus(x[:, 2 * self.event_dim : 3 * self.event_dim]), min=self.min_alpha
        )
        beta = torch.clamp(F.softplus(x[:, 3 * self.event_dim :]), min=self.min_beta)
        return {
            "loc": loc,
            "lmbda": lmbda,
            "alpha": alpha,
            "beta": beta,
        }



"""These layers are still work in progress."""

import math

import torch
import torch.nn.functional as F
from torch import Tensor, nn
from torch.nn.common_types import _size_2_t


def check_lpbnn_parameters_consistency(
    hidden_size: int, std_factor: float, num_estimators: int
) -> None:
    if hidden_size < 1:
        raise ValueError(f"hidden_size must be greater than 0. Got {hidden_size}.")
    if std_factor < 0:
        raise ValueError(f"std_factor must be greater than 0. Got {std_factor}.")
    if num_estimators < 1:
        raise ValueError(f"num_estimators must be greater than 0. Got {num_estimators}.")


def _sample(mu: Tensor, logvar: Tensor, std_factor: float) -> Tensor:
    eps = torch.randn_like(mu)
    return eps * std_factor * torch.exp(logvar * 0.5) + mu


class LPBNNLinear(nn.Module):
    __constants__ = [
        "in_features",
        "out_features",
        "num_estimators",
        "hidden_size",
    ]
    in_features: int
    out_features: int
    num_estimators: int
    bias: torch.Tensor | None

    def __init__(
        self,
        in_features: int,
        out_features: int,
        num_estimators: int,
        hidden_size: int = 32,
        std_factor: float = 1e-2,
        bias: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        """LPBNN-style linear layer.

        Args:
            in_features (int): Number of input features.
            out_features (int): Number of output features.
            num_estimators (int): Number of models to sample from.
            hidden_size (int): Size of the hidden layer. Defaults to 32.
            std_factor (float): Factor to multiply the standard deviation of the
                latent noise. Defaults to 1e-2.
            bias (bool): If ``True``, adds a learnable bias to the output.
                Defaults to ``True``.
            device (torch.device): Device on which the layer is stored.
                Defaults to ``None``.
            dtype (torch.dtype): Data type of the layer. Defaults to ``None``.

        Reference:
            `Encoding the latent posterior of Bayesian Neural Networks for
            uncertainty quantification <https://arxiv.org/abs/2012.02818>`_.
        """
        check_lpbnn_parameters_consistency(hidden_size, std_factor, num_estimators)
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.hidden_size = hidden_size
        self.std_factor = std_factor
        self.num_estimators = num_estimators

        # for the KL Loss
        self.lprior = 0

        self.linear = nn.Linear(in_features, out_features, bias=False, **factory_kwargs)
        self.alpha = nn.Parameter(
            torch.empty((num_estimators, in_features), **factory_kwargs),
            requires_grad=False,
        )
        self.gamma = nn.Parameter(torch.empty((num_estimators, out_features), **factory_kwargs))
        self.encoder = nn.Linear(in_features, self.hidden_size, **factory_kwargs)
        self.latent_mean = nn.Linear(self.hidden_size, self.hidden_size, **factory_kwargs)
        self.latent_logvar = nn.Linear(self.hidden_size, self.hidden_size, **factory_kwargs)
        self.decoder = nn.Linear(self.hidden_size, in_features, **factory_kwargs)
        self.latent_loss = torch.zeros(1, **factory_kwargs)
        if bias:
            self.bias = nn.Parameter(torch.empty((num_estimators, out_features), **factory_kwargs))
        else:
            self.register_parameter("bias", None)
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.normal_(self.alpha, mean=1.0, std=0.1)
        nn.init.normal_(self.gamma, mean=1.0, std=0.1)
        self.linear.reset_parameters()
        self.encoder.reset_parameters()
        self.decoder.reset_parameters()
        self.latent_mean.reset_parameters()
        self.latent_logvar.reset_parameters()
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.linear.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, x: Tensor) -> Tensor:
        # Draw a sample from the dist generated by the noise self.alpha
        latent = F.relu(self.encoder(self.alpha))
        latent_mean, latent_logvar = (
            self.latent_mean(latent),
            self.latent_logvar(latent),
        )
        z_latent = _sample(latent_mean, latent_logvar, self.std_factor)

        # one sample per "model" with as many features as x
        alpha_sample = self.decoder(z_latent)

        # Compute the latent loss
        if self.training:
            mse = F.mse_loss(alpha_sample, self.alpha)
            kld = -0.5 * torch.sum(1 + latent_logvar - latent_mean**2 - torch.exp(latent_logvar))
            # For the KL Loss
            self.lvposterior = mse + kld

        # Compute the output
        num_examples_per_model = int(x.size(0) / self.num_estimators)
        alpha = alpha_sample.repeat((num_examples_per_model, 1))
        gamma = self.gamma.repeat((num_examples_per_model, 1))
        out = self.linear(x * alpha) * gamma

        if self.bias is not None:
            bias = self.bias.repeat((num_examples_per_model, 1))
            out += bias
        return out

    def extra_repr(self) -> str:
        return (
            f"in_features={self.in_features}, "
            f"out_features={self.out_features}, "
            f"num_estimators={self.num_estimators}, "
            f"hidden_size={self.hidden_size}, bias={self.bias is not None}"
        )


class LPBNNConv2d(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        num_estimators: int,
        kernel_size: _size_2_t,
        stride: _size_2_t = 1,
        padding: str | _size_2_t = 0,
        groups: int = 1,
        hidden_size: int = 32,
        std_factor: float = 1e-2,
        gamma: bool = True,
        bias: bool = True,
        padding_mode: str = "zeros",
        device=None,
        dtype=None,
    ):
        """LPBNN-style 2D convolutional layer.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            num_estimators (int): Number of models to sample from.
            kernel_size (int or tuple): Size of the convolving kernel.
            stride (int or tuple, optional): Stride of the convolution. Default: 1.
            padding (int or tuple, optional): Zero-padding added to both sides of the input. Default: 0.
            groups (int, optional): Number of blocked connections from input channels to output channels. Default: 1.
            hidden_size (int): Size of the hidden layer. Defaults to 32.
            std_factor (float): Factor to multiply the standard deviation of the
                latent noise. Defaults to 1e-2.
            gamma (bool): If ``True``, adds a learnable gamma to the output.
                Defaults to ``True``.
            bias (bool): If ``True``, adds a learnable bias to the output.
                Defaults to ``True``.
            padding_mode (str): 'zeros', 'reflect', 'replicate' or 'circular'. Default: 'zeros'.
            device (torch.device): Device on which the layer is stored.
                Defaults to ``None``.
            dtype (torch.dtype): Data type of the layer. Defaults to ``None``.

        Reference:
            `Encoding the latent posterior of Bayesian Neural Networks for
            uncertainty quantification <https://arxiv.org/abs/2012.02818>`_.
        """
        check_lpbnn_parameters_consistency(hidden_size, std_factor, num_estimators)
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.hidden_size = hidden_size
        self.std_factor = std_factor
        self.num_estimators = num_estimators

        # for the KL Loss
        self.lprior = 0

        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            groups=groups,
            bias=False,
            padding_mode=padding_mode,
        )
        self.alpha = nn.Parameter(
            torch.empty(num_estimators, in_channels, **factory_kwargs),
            requires_grad=False,
        )

        self.encoder = nn.Linear(in_channels, self.hidden_size, **factory_kwargs)
        self.decoder = nn.Linear(self.hidden_size, in_channels, **factory_kwargs)
        self.latent_mean = nn.Linear(self.hidden_size, self.hidden_size, **factory_kwargs)
        self.latent_logvar = nn.Linear(self.hidden_size, self.hidden_size, **factory_kwargs)

        self.latent_loss = torch.zeros(1, **factory_kwargs)
        if gamma:
            self.gamma = nn.Parameter(torch.empty((num_estimators, out_channels), **factory_kwargs))
        else:
            self.register_parameter("gamma", None)

        if bias:
            self.bias = nn.Parameter(torch.empty((num_estimators, out_channels), **factory_kwargs))
        else:
            self.register_parameter("bias", None)
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.normal_(self.alpha, mean=1.0, std=0.1)
        if self.gamma is not None:
            nn.init.normal_(self.gamma, mean=1.0, std=0.1)
        self.conv.reset_parameters()
        self.encoder.reset_parameters()
        self.decoder.reset_parameters()
        self.latent_mean.reset_parameters()
        self.latent_logvar.reset_parameters()
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.conv.weight)
            bound = 1 / math.sqrt(fan_in)
            nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, x: Tensor) -> Tensor:
        # Draw a sample from the dist generated by the latent noise self.alpha
        latent = F.relu(self.encoder(self.alpha))
        latent_mean, latent_logvar = (
            self.latent_mean(latent),
            self.latent_logvar(latent),
        )
        z_latent = _sample(latent_mean, latent_logvar, self.std_factor)

        # one sample per "model" with as many features as x
        alpha_sample = self.decoder(z_latent)

        # Compute the latent loss
        if self.training:
            mse = F.mse_loss(alpha_sample, self.alpha)
            kld = -0.5 * torch.sum(1 + latent_logvar - latent_mean.pow(2) - latent_logvar.exp())
            # for the KL Loss
            self.lvposterior = mse + kld

        num_examples_per_model = int(x.size(0) / self.num_estimators)

        # Compute the output
        alpha = alpha_sample.repeat((num_examples_per_model, 1)).unsqueeze(-1).unsqueeze(-1)
        if self.gamma is not None:
            gamma = self.gamma.repeat((num_examples_per_model, 1)).unsqueeze(-1).unsqueeze(-1)
            out = self.conv(x * alpha) * gamma
        else:
            out = self.conv(x * alpha)

        if self.bias is not None:
            bias = self.bias.repeat((num_examples_per_model, 1)).unsqueeze(-1).unsqueeze(-1)
            out += bias

        return out

    def extra_repr(self) -> str:
        return (
            f"in_channels={self.in_channels}, "
            f"out_channels={self.out_channels}, "
            f"num_estimators={self.num_estimators}, "
            f"hidden_size={self.hidden_size}, "
            f"gamma={self.gamma is not None}, "
            f"bias={self.bias is not None}"
        )



from typing import Any, Literal

import torch
import torch.nn.functional as F
from torch import Tensor
from torchmetrics import Metric
from torchmetrics.utilities.data import dim_zero_cat


class CategoricalNLL(Metric):
    is_differentiable = False
    higher_is_better = False
    full_state_update = False

    def __init__(
        self,
        reduction: Literal["mean", "sum", "none", None] = "mean",
        **kwargs: Any,
    ) -> None:
        """The Negative Log Likelihood Metric.

        Args:
            reduction (str, optional): Determines how to reduce over the
                :math:`B`/batch dimension:

                - ``'mean'`` [default]: Averages score across samples
                - ``'sum'``: Sum score across samples
                - ``'none'`` or ``None``: Returns score per sample

            kwargs: Additional keyword arguments, see `Advanced metric settings
                <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.

        Inputs:
            - :attr:`probs`: :math:`(B, C)`
            - :attr:`target`: :math:`(B)`

            where :math:`B` is the batch size and :math:`C` is the number of
            classes.

        Warning:
            Make sure that the probabilities in :attr:`probs` are normalized to sum
            to one.

        Raises:
            ValueError:
                If :attr:`reduction` is not one of ``'mean'``, ``'sum'``,
                ``'none'`` or ``None``.
        """
        super().__init__(**kwargs)

        allowed_reduction = ("sum", "mean", "none", None)
        if reduction not in allowed_reduction:
            raise ValueError(
                "Expected argument `reduction` to be one of ",
                f"{allowed_reduction} but got {reduction}",
            )

        self.reduction = reduction

        if self.reduction in ["mean", "sum"]:
            self.add_state(
                "values",
                default=torch.tensor(0.0),
                dist_reduce_fx="sum",
            )
        else:
            self.add_state("values", default=[], dist_reduce_fx="cat")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, probs: Tensor, target: Tensor) -> None:
        """Update state with prediction probabilities and targets.

        Args:
            probs (Tensor): Probabilities from the model.
            target (Tensor): Ground truth labels.
        """
        if self.reduction is None or self.reduction == "none":
            self.values.append(F.nll_loss(torch.log(probs), target, reduction="none"))
        else:
            self.values += F.nll_loss(torch.log(probs), target, reduction="sum")
            self.total += target.size(0)

    def compute(self) -> Tensor:
        """Computes NLL based on inputs passed in to ``update`` previously."""
        values = dim_zero_cat(self.values)

        if self.reduction == "sum":
            return values.sum(dim=-1)
        if self.reduction == "mean":
            return values.sum(dim=-1) / self.total
        # reduction is None or "none"
        return values



import pytest
import torch

from torch_uncertainty.models.resnet import (
    batched_resnet,
    lpbnn_resnet,
    masked_resnet,
    mimo_resnet,
    packed_resnet,
    resnet,
)
from torch_uncertainty.models.resnet.utils import get_resnet_num_blocks


class TestResnet:
    """Testing the ResNet classes."""

    def test_main(self):
        resnet(1, 10, arch=18, conv_bias=True, style="cifar")
        model = resnet(1, 10, arch=50, style="cifar")
        with torch.no_grad():
            model(torch.randn(1, 1, 32, 32))
            model.feats_forward(torch.randn(1, 1, 32, 32))

        get_resnet_num_blocks(44)
        get_resnet_num_blocks(56)
        get_resnet_num_blocks(110)
        get_resnet_num_blocks(1202)

    def test_mc_dropout(self):
        resnet(1, 10, arch=20, conv_bias=False, style="cifar")
        model = resnet(1, 10, arch=50).eval()
        with torch.no_grad():
            model(torch.randn(1, 1, 32, 32))

    def test_error(self):
        with pytest.raises(ValueError):
            resnet(1, 10, arch=20, style="test")


class TestPackedResnet:
    """Testing the ResNet packed class."""

    def test_main(self):
        model = packed_resnet(1, 10, 20, 2, 2, 1)
        model = packed_resnet(1, 10, 152, 2, 2, 1)
        assert model.check_config({"alpha": 2, "gamma": 1, "groups": 1, "num_estimators": 2})
        assert not model.check_config({"alpha": 1, "gamma": 1, "groups": 1, "num_estimators": 2})

    def test_error(self):
        with pytest.raises(ValueError):
            packed_resnet(1, 10, 20, 2, 2, 1, style="test")


class TestMaskedResnet:
    """Testing the ResNet masked class."""

    def test_main(self):
        model = masked_resnet(1, 10, 20, 2, 2)
        with torch.no_grad():
            model(torch.randn(1, 1, 32, 32))

    def test_error(self):
        with pytest.raises(ValueError):
            masked_resnet(1, 10, 20, 2, 2, style="test")


class TestBatchedResnet:
    """Testing the ResNet batched class."""

    def test_main(self):
        model = batched_resnet(1, 10, 20, 2, conv_bias=True)
        with torch.no_grad():
            model(torch.randn(1, 1, 32, 32))

    def test_error(self):
        with pytest.raises(ValueError):
            batched_resnet(1, 10, 20, 2, style="test")


class TestLPBNNResnet:
    """Testing the ResNet LPBNN class."""

    def test_main(self):
        model = lpbnn_resnet(1, 10, 20, 2, conv_bias=True)
        with torch.no_grad():
            model(torch.randn(1, 1, 32, 32))
        model = lpbnn_resnet(1, 10, 50, 2, conv_bias=False, style="cifar")
        with torch.no_grad():
            model(torch.randn(1, 1, 32, 32))

    def test_error(self):
        with pytest.raises(ValueError):
            lpbnn_resnet(1, 10, 20, 2, style="test")
        with pytest.raises(ValueError, match="Unknown ResNet architecture. Got"):
            lpbnn_resnet(1, 10, 42, 2, style="test")


class TestMIMOResnet:
    """Testing the ResNet MIMO class."""

    def test_main(self):
        model = mimo_resnet(1, 10, 34, 2, style="cifar", conv_bias=False)
        model.train()
        model(torch.rand((2, 1, 28, 28)))

    def test_error(self):
        with pytest.raises(ValueError):
            mimo_resnet(1, 10, 101, 2, style="test")



"""Modified from https://github.com/nikitadurasov/masksembles/."""

from typing import Any

import numpy as np
import torch
from torch import Tensor, nn
from torch.nn.common_types import _size_2_t


def _generate_masks(m: int, n: int, s: float) -> np.ndarray:
    """Generates set of binary masks with properties defined by n, m, s params.
    Results of this function are stochastic, that is, calls with the same sets
    of arguments might generate outputs of different shapes. Check
    generate_masks and generation_wrapper function for more deterministic
    behaviour.

    Args:
        m (int): Number of ones in each mask.
        n (int): Number of masks in the set.
        s (float): Scale param controls overlap of generated masks.

    Returns:
        np.ndarray: Matrix of binary vectors.
    """
    rng = np.random.default_rng()
    total_positions = int(m * s)
    masks = []

    for _ in range(n):
        new_vector = np.zeros([total_positions])
        idx = rng.choice(range(total_positions), m, replace=False)
        new_vector[idx] = 1
        masks.append(new_vector)

    masks = np.array(masks)
    # drop useless positions
    return masks[:, ~np.all(masks == 0, axis=0)]


def generate_masks(m: int, n: int, s: float) -> np.ndarray:
    """Generates set of binary masks with properties defined by n, m, s params
    Resulting masks are required to have fixed features size.
    Since process of masks generation is stochastic therefore function
    evaluates _generate_masks multiple times till expected size is acquired.

    Args:
        m (int): number of ones in each mask
        n (int): number of masks in the set
        s (float): scale param controls overlap of generated masks

    Returns:
        np.ndarray: matrix of binary vectors
    """
    masks = _generate_masks(m, n, s)
    # hardcoded formula for expected size, check reference
    expected_size = int(m * s * (1 - (1 - 1 / s) ** n))
    while masks.shape[1] != expected_size:
        masks = _generate_masks(m, n, s)
    return masks


def generation_wrapper(c: int, n: int, scale: float) -> np.ndarray:
    """Generates set of binary masks with properties defined by c, n, scale
    params. Allows to generate masks sets with predefined features number c.
    Particularly convenient to use in torch-like layers where one need to
    define shapes inputs tensors beforehand.

    Args:
        c (int): number of channels in generated masks.
        n (int): number of masks in the set.
        scale (float): scale param controls overlap of generated masks.

    Raises:
        ValueError: If :attr:`c` < 10.
        ValueError: If :attr:`s` > 0.6.

    Returns:
        np.ndarray: matrix of binary vectors
    """
    if c < 10:
        raise ValueError(
            "Masksembles approach couldn't be used in such setups where "
            "number of channels is less then 10. Current value is "
            f"(channels={c})."
            "Please increase number of features in your layer or remove this "
            "particular instance of Masksembles from your architecture."
        )

    if scale > 6.0:
        raise ValueError(
            "Masksembles approach couldn't be used in such setups where "
            "scale parameter is larger then 6. Current value is  "
            f"(scale={scale})."
        )

    # inverse formula for number of active features in masks
    active_features = int(int(c) / (scale * (1 - (1 - 1 / scale) ** n)))

    # Use binary search to find the correct value of the scale
    masks = generate_masks(active_features, n, scale)
    up = 4 * scale
    down = max(0.2 * scale, 1.0)
    s = (down + up) / 2
    im_s = -1
    while im_s != c:
        masks = generate_masks(active_features, n, s)
        im_s = masks.shape[-1]
        if im_s < c:
            down = s
            s = (down + up) / 2
        elif im_s > c:
            up = s
            s = (down + up) / 2

    return masks


class Mask1d(nn.Module):
    def __init__(self, channels: int, num_masks: int, scale: float, **factory_kwargs) -> None:
        super().__init__()
        self.num_masks = num_masks

        masks = generation_wrapper(channels, num_masks, scale)
        masks = torch.from_numpy(masks)
        self.masks = torch.nn.Parameter(masks, requires_grad=False).to(
            device=factory_kwargs["device"]
        )

    def forward(self, inputs: Tensor) -> Tensor:
        batch = inputs.shape[0]
        x = torch.split(inputs.unsqueeze(1), batch // self.num_masks, dim=0)
        x = torch.cat(x, dim=1).permute([1, 0, 2])
        x = x * self.masks.unsqueeze(1)
        x = torch.cat(torch.split(x, 1, dim=0), dim=1)
        return torch.as_tensor(x, dtype=inputs.dtype).squeeze(0)


class Mask2d(nn.Module):
    def __init__(self, channels: int, num_masks: int, scale: float, **factory_kwargs) -> None:
        super().__init__()
        self.num_masks = num_masks

        masks = generation_wrapper(channels, num_masks, scale)
        masks = torch.from_numpy(masks)
        self.masks = torch.nn.Parameter(masks, requires_grad=False).to(
            device=factory_kwargs["device"]
        )

    def forward(self, inputs: Tensor) -> Tensor:
        batch = inputs.shape[0]
        x = torch.split(inputs.unsqueeze(1), batch // self.num_masks, dim=0)
        x = torch.cat(x, dim=1).permute([1, 0, 2, 3, 4])
        x = x * self.masks.unsqueeze(1).unsqueeze(-1).unsqueeze(-1)
        x = torch.cat(torch.split(x, 1, dim=0), dim=1)
        return torch.as_tensor(x, dtype=inputs.dtype).squeeze(0)


class MaskedLinear(nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        num_estimators: int,
        scale: float,
        bias: bool = True,
        device: Any | None = None,
        dtype: Any | None = None,
    ) -> None:
        r"""Masksembles-style Linear layer.

        This layer computes fully-connected operation for a given number of
        estimators (:attr:`num_estimators`) with a given :attr:`scale`.

        Args:
            in_features (int): Number of input features of the linear layer.
            out_features (int): Number of channels produced by the linear layer.
            num_estimators (int): The number of estimators grouped in the layer.
            scale (float): The scale parameter for the masks.
            bias (bool, optional): It ``True``, adds a learnable bias to the
                output. Defaults to ``True``.
            groups (int, optional): Number of blocked connections from input
                channels to output channels. Defaults to ``1``.
            device (Any, optional): The desired device of returned tensor.
                Defaults to ``None``.
            dtype (Any, optional): The desired data type of returned tensor.
                Defaults to ``None``.

        Warning:
            Be sure to apply a repeat on the batch at the start of the training
            if you use `MaskedLinear`.

        Reference:
            `Masksembles for Uncertainty Estimation`, Nikita Durasov, Timur
            Bagautdinov, Pierre Baque, Pascal Fua.
        """
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        if scale is None:
            raise ValueError("You must specify the value of the arg. `scale`")
        if scale < 1:
            raise ValueError(f"Attribute `scale` should be >= 1, not {scale}.")

        self.mask = Mask1d(in_features, num_masks=num_estimators, scale=scale, **factory_kwargs)
        self.linear = nn.Linear(
            in_features=in_features,
            out_features=out_features,
            bias=bias,
            **factory_kwargs,
        )

    def forward(self, inputs: Tensor) -> Tensor:
        return self.linear(self.mask(inputs))


class MaskedConv2d(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        num_estimators: int,
        scale: float,
        stride: _size_2_t = 1,
        padding: str | _size_2_t = 0,
        dilation: _size_2_t = 1,
        groups: int = 1,
        bias: bool = True,
        device: Any | None = None,
        dtype: Any | None = None,
    ) -> None:
        r"""Masksembles-style Conv2d layer.

        Args:
            in_channels (int): Number of channels in the input image.
            out_channels (int): Number of channels produced by the convolution.
            kernel_size (int or tuple): Size of the convolving kernel.
            num_estimators (int): Number of estimators in the ensemble.
            scale (float): The scale parameter for the masks.
            stride (int or tuple, optional): Stride of the convolution.
                Defaults to ``1``.
            padding (int, tuple or str, optional): Padding added to all four sides
                of the input. Defaults to ``0``.
            dilation (int or tuple, optional): Spacing between kernel elements.
                Defaults to ``1``.
            groups (int, optional): Number of blocked connexions from input
                channels to output channels for each estimator. Defaults to ``1``.
            bias (bool, optional): If ``True``, adds a learnable bias to the
                output. Defaults to ``True``.
            device (Any, optional): The desired device of returned tensor.
                Defaults to ``None``.
            dtype (Any, optional): The desired data type of returned tensor.
                Defaults to ``None``.

        Warning:
            Be sure to apply a repeat on the batch at the start of the training
            if you use `MaskedConv2d`.

        Reference:
            `Masksembles for Uncertainty Estimation`, Nikita Durasov, Timur
            Bagautdinov, Pierre Baque, Pascal Fua.
        """
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        if scale is None:
            raise ValueError("You must specify the value of the arg. `scale`")
        if scale < 1:
            raise ValueError(f"Attribute `scale` should be >= 1, not {scale}.")

        self.mask = Mask2d(in_channels, num_masks=num_estimators, scale=scale, **factory_kwargs)
        self.conv = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode="zeros",
            **factory_kwargs,
        )

    def forward(self, inputs: Tensor) -> Tensor:
        return self.conv(self.mask(inputs))



import torch
from torch import Tensor
from torchmetrics import Metric
from torchmetrics.utilities.data import dim_zero_cat


class ThresholdAccuracy(Metric):
    def __init__(self, power: int, lmbda: float = 1.25, **kwargs) -> None:
        r"""The Threshold Accuracy metric, a.k.a. d1, d2, d3.

        Args:
            power: The power to raise the threshold to. Often in [1, 2, 3].
            lmbda: The threshold to compare the max of ratio of predictions
                to targets and its inverse to. Defaults to 1.25.
            kwargs: Additional keyword arguments, see `Advanced metric settings
                <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.
        """
        super().__init__(**kwargs)
        if power < 0:
            raise ValueError(f"Power must be greater than or equal to 0. Got {power}.")
        self.power = power
        if lmbda < 1:
            raise ValueError(f"Lambda must be greater than or equal to 1. Got {lmbda}.")
        self.lmbda = lmbda
        self.add_state("values", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, preds: Tensor, target: Tensor) -> None:
        """Update state with predictions and targets."""
        self.values += torch.sum(torch.max(preds / target, target / preds) < self.lmbda**self.power)
        self.total += target.size(0)

    def compute(self) -> Tensor:
        """Compute the Threshold Accuracy."""
        values = dim_zero_cat(self.values)
        return values / self.total



from typing import Any

from einops import rearrange
from torch import Tensor, nn

from torch_uncertainty.layers.packed import PackedConv2d, PackedLinear


class VGG(nn.Module):
    def __init__(
        self,
        vgg_cfg: list[str | int],
        in_channels: int,
        num_classes: int,
        linear_layer: type[nn.Module],
        conv2d_layer: type[nn.Module],
        norm: type[nn.Module],
        groups: int,
        dropout_rate: float,
        style: str,
        **model_kwargs: Any,
    ) -> None:
        super().__init__()

        self.in_channels = in_channels
        self.linear_layer = linear_layer
        self.conv2d_layer = conv2d_layer
        self.norm = norm
        self.groups = groups
        self.dropout_rate = dropout_rate

        if self.conv2d_layer == PackedConv2d:
            self.num_estimators = model_kwargs.get("num_estimators")
        else:
            self.num_estimators = model_kwargs.pop("num_estimators")
        self.last_layer_dropout = model_kwargs.pop("last_layer_dropout", False)
        self.model_kwargs = model_kwargs

        self.features = self._make_layers(vgg_cfg)

        if style == "imagenet":
            self.avgpool = nn.AdaptiveAvgPool2d((7, 7))
            kernel_surface = 7 * 7
        else:
            self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
            kernel_surface = 1

        if self.linear_layer == PackedLinear:
            last_linear = linear_layer(4096, num_classes, last=True, **model_kwargs)
        else:
            last_linear = linear_layer(4096, num_classes, **model_kwargs)

        self.cls_head_layers = [
            linear_layer(512 * kernel_surface, 4096, **model_kwargs),
            nn.ReLU(True),
            nn.Dropout(p=dropout_rate),
            linear_layer(4096, 4096, **model_kwargs),
            nn.ReLU(True),
            nn.Dropout(p=dropout_rate),
            last_linear,
        ]

        self.flatten = nn.Flatten(1)
        self.cls_head = nn.Sequential(*self.cls_head_layers)
        self._init_weights()

    def _init_weights(self) -> None:
        for m in self.modules():
            if isinstance(m, nn.Conv2d | PackedConv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
                if m.bias is not None:  # coverage: ignore
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):  # coverage: ignore
                nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear | PackedLinear):
                nn.init.normal_(m.weight, 0, 0.01)
                if m.bias is not None:  # coverage: ignore
                    nn.init.constant_(m.bias, 0)

    def _make_layers(self, cfg: list) -> nn.Sequential:
        layers: list[nn.Module] = []
        in_channels = self.in_channels
        for i, v in enumerate(cfg):
            if v == "M":
                layers.append(nn.MaxPool2d(kernel_size=2, stride=2))
            else:
                if i == 0 and self.conv2d_layer == PackedConv2d:
                    conv2d = self.conv2d_layer(
                        in_channels,
                        v,
                        kernel_size=3,
                        padding=1,
                        groups=self.groups,
                        first=True,
                        **self.model_kwargs,
                    )
                else:
                    conv2d = self.conv2d_layer(
                        in_channels,
                        v,
                        kernel_size=3,
                        padding=1,
                        groups=self.groups if i != 0 else 1,
                        **self.model_kwargs,
                    )
                layers.extend([conv2d, self.norm(v), nn.ReLU(inplace=True)])
                in_channels = v
        return nn.Sequential(*layers)

    def feats_forward(self, x: Tensor) -> Tensor:
        x = self.features(x)

        if self.linear_layer == PackedLinear:
            x = rearrange(
                x,
                "e (m c) h w -> (m e) c h w",
                m=self.model_kwargs["num_estimators"],
            )
        x = self.avgpool(x)
        return self.flatten(x)

    def forward(self, x: Tensor) -> Tensor:
        return self.cls_head(self.feats_forward(x))


def _vgg(
    vgg_cfg: list[str | int],
    in_channels: int,
    num_classes: int,
    linear_layer: type[nn.Module] = nn.Linear,
    conv2d_layer: type[nn.Module] = nn.Conv2d,
    norm: type[nn.Module] = nn.Identity,
    groups: int = 1,
    dropout_rate: float = 0.5,
    style: str = "imagenet",
    **model_kwargs: Any,
) -> VGG:
    return VGG(
        vgg_cfg,
        in_channels,
        num_classes,
        linear_layer,
        conv2d_layer,
        norm,
        groups,
        dropout_rate,
        style,
        **model_kwargs,
    )



import copy

import torch
from torch import Tensor, nn
from torch.utils.data import DataLoader, Dataset

from torch_uncertainty.layers.bayesian.abnn import BatchNormAdapter2d
from torch_uncertainty.models import deep_ensembles
from torch_uncertainty.optim_recipes import optim_abnn
from torch_uncertainty.routines import ClassificationRoutine
from torch_uncertainty.utils import TUTrainer

from .abstract import PostProcessing


class ABNN(PostProcessing):
    def __init__(
        self,
        num_classes: int,
        random_prior: float,
        alpha: float,
        num_models: int,
        num_samples: int,
        base_lr: float,
        device: torch.device | str,
        max_epochs: int = 5,
        use_original_model: bool = True,
        batch_size: int = 128,
        precision: str = "32",
        model: nn.Module | None = None,
    ):
        """ABNN post-processing.

        Args:
            num_classes (int): Number of classes of the inner model.
            random_prior (float): Random prior specializing estimators on
                certain classes.
            alpha (float): Alpha value for ABNN to control the diversity of
                the predictions.
            num_models (int): Number of stochastic models.
            num_samples (int): Number of samples per model.
            base_lr (float): Base learning rate.
            device (torch.device): Device to use.
            max_epochs (int, optional): Number of training epochs. Defaults
                to 5.
            use_original_model (bool, optional): Use original model during
                evaluation. Defaults to True.
            batch_size (int, optional): Batch size for the training of ABNN.
                Defaults to 128.
            precision (str, optional): Machine precision for training & eval.
                Defaults to "32".
            model (nn.Module | None, optional): Model to use. Defaults to None.

        Reference:

        """
        super().__init__(model)
        _abnn_checks(
            num_classes=num_classes,
            random_prior=random_prior,
            alpha=alpha,
            max_epochs=max_epochs,
            num_models=num_models,
            num_samples=num_samples,
            base_lr=base_lr,
            batch_size=batch_size,
        )
        self.num_classes = num_classes
        self.alpha = alpha
        self.base_lr = base_lr
        self.num_models = num_models
        self.num_samples = num_samples
        self.total_models = num_models + int(use_original_model)
        self.use_original_model = use_original_model
        self.max_epochs = max_epochs

        self.batch_size = batch_size
        self.precision = precision
        self.device = device

        self.final_model = None

        # Build random prior
        num_rp_classes = int(num_classes**0.5)
        self.weights = []
        for _ in range(num_models):
            weight = torch.ones([num_classes])
            weight[torch.randperm(num_classes)[:num_rp_classes]] += random_prior - 1
            self.weights.append(weight)

    def fit(self, dataset: Dataset) -> None:
        if self.model is None:
            raise ValueError("Model must be set before fitting.")
        dl = DataLoader(dataset, batch_size=self.batch_size, shuffle=True)

        source_model = copy.deepcopy(self.model)
        _replace_bn_layers(source_model, self.alpha)

        models = [copy.deepcopy(source_model) for _ in range(self.num_models)]

        baselines = [
            ClassificationRoutine(
                num_classes=self.num_classes,
                model=mod,
                loss=nn.CrossEntropyLoss(weight=self.weights[i].to(device=self.device)),
                optim_recipe=optim_abnn(mod, lr=self.base_lr),
                eval_ood=True,
            )
            for i, mod in enumerate(models)
        ]

        for baseline in baselines:
            trainer = TUTrainer(
                max_epochs=self.max_epochs,
                accelerator=self.device,
                enable_progress_bar=False,
                precision=self.precision,
                enable_checkpointing=False,
                logger=None,
                enable_model_summary=False,
            )
            trainer.fit(model=baseline, train_dataloaders=dl)

        final_models = (
            [copy.deepcopy(source_model) for _ in range(self.num_samples)]
            if self.use_original_model
            else []
        )
        for baseline in baselines:
            model = copy.deepcopy(source_model)
            model.load_state_dict(baseline.model.state_dict())
            final_models.extend([copy.deepcopy(model) for _ in range(self.num_samples)])

        self.final_model = deep_ensembles(final_models)

    def forward(
        self,
        x: Tensor,
    ) -> Tensor:
        if self.final_model is not None:
            return self.final_model(x)
        if self.model is not None:
            return self.model(x)
        raise ValueError("Model must be set before calling forward.")


def _abnn_checks(
    num_classes,
    random_prior,
    alpha,
    max_epochs,
    num_models,
    num_samples,
    base_lr,
    batch_size,
) -> None:
    if random_prior < 0:
        raise ValueError(f"random_prior must be greater than 0. Got {random_prior}.")
    if batch_size < 1:
        raise ValueError(f"batch_size must be greater than 0. Got {batch_size}.")
    if max_epochs < 1:
        raise ValueError(f"epoch must be greater than 0. Got {max_epochs}.")
    if num_models < 1:
        raise ValueError(f"num_models must be greater than 0. Got {num_models}.")
    if num_samples < 1:
        raise ValueError(f"num_samples must be greater than 0. Got {num_samples}.")
    if alpha < 0:
        raise ValueError(f"alpha must be greater than 0. Got {alpha}.")
    if base_lr < 0:
        raise ValueError(f"base_lr must be greater than 0. Got {base_lr}.")
    if num_classes < 1:
        raise ValueError(f"num_classes must be greater than 0. Got {num_classes}.")


def _replace_bn_layers(model: nn.Module, alpha: float) -> None:
    """Recursively replace batch normalization layers with ABNN layers.

    Args:
        model (nn.Module): Model to replace batch normalization layers.
        alpha (float): Alpha value for ABNN.
    """
    for name, module in model.named_children():
        if len(list(module.children())) > 0:
            _replace_bn_layers(module, alpha)
        if isinstance(module, nn.BatchNorm2d) and module.track_running_stats:
            num_channels = module.num_features
            new_module = BatchNormAdapter2d(num_channels, alpha=alpha)
            new_module.running_mean = module.running_mean
            new_module.running_var = module.running_var
            new_module.num_batches_tracked = module.num_batches_tracked

            new_module.weight.data = module.weight.data
            new_module.bias.data = module.bias.data
            setattr(model, name, new_module)
        else:
            _replace_bn_layers(module, alpha)



import pytest
import torch

from torch_uncertainty.metrics import VariationRatio


@pytest.fixture()
def disagreement_probas_3est() -> torch.Tensor:
    """Return a vector with mean entropy ~ln(2) and entropy of mean =0."""
    return torch.as_tensor([[[0.2, 0.8]], [[0.7, 0.3]], [[0.6, 0.4]]])


@pytest.fixture()
def agreement_probas() -> torch.Tensor:
    return torch.as_tensor([[[0.9, 0.1]], [[0.9, 0.1]]])


@pytest.fixture()
def agreement_probas_3est() -> torch.Tensor:
    """Return a vector with mean entropy ~ln(2) and entropy of mean =0."""
    return torch.as_tensor([[[0.2, 0.8]], [[0.3, 0.7]], [[0.4, 0.6]]])


class TestVariationRatio:
    """Testing the VariationRatio metric class."""

    def test_compute_agreement(self, agreement_probas: torch.Tensor):
        metric = VariationRatio(probabilistic=True)
        metric.update(agreement_probas)
        res = metric.compute()
        assert res == pytest.approx(0.1, 1e-6)

        metric = VariationRatio(probabilistic=False)
        metric.update(agreement_probas)
        res = metric.compute()
        assert res == 0.0

        metric = VariationRatio(probabilistic=False, reduction="none")
        metric.update(agreement_probas)
        res = metric.compute()
        assert res == 0.0

    def test_compute_disagreement(
        self,
        agreement_probas_3est: torch.Tensor,
        disagreement_probas_3est: torch.Tensor,
    ):
        metric = VariationRatio(probabilistic=True, reduction="sum")
        metric.update(disagreement_probas_3est)
        metric.update(agreement_probas_3est)
        res = metric.compute()
        assert res == pytest.approx(0.8, 1e-6)

    def test_bad_argument(self):
        with pytest.raises(ValueError):
            _ = VariationRatio(reduction="geometric_mean")



import pytest
import torch
from torch import nn, optim

from torch_uncertainty.layers.bayesian import BayesLinear
from torch_uncertainty.layers.distributions import NormalLinear
from torch_uncertainty.losses import DistributionNLLLoss, ELBOLoss
from torch_uncertainty.routines import RegressionRoutine


class TestELBOLoss:
    """Testing the ELBOLoss class."""

    def test_main(self):
        model = BayesLinear(1, 1)
        criterion = nn.BCEWithLogitsLoss()
        loss = ELBOLoss(model, criterion, kl_weight=1e-5, num_samples=1)
        loss(model(torch.randn(1, 1)), torch.randn(1, 1))

        model = nn.Linear(1, 1)
        criterion = nn.BCEWithLogitsLoss()

        ELBOLoss(None, criterion, kl_weight=1e-5, num_samples=1)
        loss = ELBOLoss(model, criterion, kl_weight=1e-5, num_samples=1)
        loss(model(torch.randn(1, 1)), torch.randn(1, 1))

    def test_prob_regression_training_step(self):
        model = NormalLinear(BayesLinear, event_dim=4, in_features=10)
        criterion = DistributionNLLLoss()
        loss = ELBOLoss(model, criterion, kl_weight=1e-5, num_samples=3, dist_family="normal")

        routine = RegressionRoutine(
            output_dim=1,
            model=model,
            loss=loss,
            dist_family="normal",
            optim_recipe=optim.Adam(
                model.parameters(),
                lr=5e-4,
                weight_decay=0,
            ),
        )
        inputs = torch.randn(1, 10)
        targets = torch.randn(1, 4)
        routine.training_step((inputs, targets))

    def test_training_step(self):
        model = BayesLinear(10, 4)
        criterion = nn.MSELoss()
        loss = ELBOLoss(model, criterion, kl_weight=1 / 50000, num_samples=3)

        routine = RegressionRoutine(
            output_dim=4,
            model=model,
            loss=loss,
            optim_recipe=optim.Adam(
                model.parameters(),
                lr=5e-4,
                weight_decay=0,
            ),
        )

        inputs = torch.randn(1, 10)
        targets = torch.randn(1, 4)
        routine.training_step((inputs, targets))

    def test_failures(self):
        model = BayesLinear(1, 1)
        criterion = nn.BCEWithLogitsLoss()

        with pytest.raises(TypeError, match="The inner_loss should be an instance of a class."):
            ELBOLoss(model, nn.BCEWithLogitsLoss, kl_weight=1, num_samples=1)

        with pytest.raises(ValueError, match="The KL weight should be non-negative. Got "):
            ELBOLoss(model, criterion, kl_weight=-1, num_samples=1)

        with pytest.raises(
            ValueError,
            match="The number of samples should not be lower than 1.",
        ):
            ELBOLoss(model, criterion, kl_weight=1, num_samples=-1)

        with pytest.raises(TypeError, match="The number of samples should be an integer. "):
            ELBOLoss(model, criterion, kl_weight=1e-5, num_samples=1.5)



# ruff: noqa: F401
from .cifar10 import CIFAR10DataModule
from .cifar100 import CIFAR100DataModule
from .imagenet import ImageNetDataModule
from .mnist import MNISTDataModule
from .tiny_imagenet import TinyImageNetDataModule
from .uci import (
    BankMarketingDataModule,
    DOTA2GamesDataModule,
    HTRU2DataModule,
    OnlineShoppersDataModule,
    SpamBaseDataModule,
)



import torch
from torch.utils.data import TensorDataset


class Cubic(TensorDataset):
    """A dataset of samples drawn from the cube fct. with homoscedastic noise.

    Args:
        lower_bound (float, optional): Lower bound of the samples. Defaults to
            ``-4.0``.
        upper_bound (float, optional): Upper bound of the samples. Defaults to
            ``4.0``.
        num_samples (int, optional): Number of samples. Defaults to ``5000``.
        noise_mean (float, optional): Mean of the noise. Defaults to ``0.0``.
        noise_std (float, optional): Standard deviation of the noise. Defaults
            to ``3.0``.
    """

    def __init__(
        self,
        lower_bound: float = -4.0,
        upper_bound: float = 4.0,
        num_samples: int = 5000,
        noise_mean: float = 0.0,
        noise_std: float = 3.0,
    ) -> None:
        noise = (noise_mean, noise_std)

        samples = torch.linspace(lower_bound, upper_bound, num_samples).unsqueeze(-1)
        targets = samples**3 + torch.normal(*noise, size=samples.size())
        super().__init__(samples, targets.squeeze(-1))



import pytest
import torch

from torch_uncertainty.metrics import Disagreement


@pytest.fixture()
def disagreement_probas() -> torch.Tensor:
    return torch.as_tensor([[[0.0, 1.0], [1.0, 0.0]]])


@pytest.fixture()
def agreement_probas() -> torch.Tensor:
    return torch.as_tensor([[[1.0, 0.0], [1.0, 0.0]]])


@pytest.fixture()
def disagreement_probas_3() -> torch.Tensor:
    return torch.as_tensor([[[0.0, 1.0], [0.0, 1.0], [1.0, 0.0]]])


class TestDisagreement:
    """Testing the Disagreement metric class."""

    def test_compute_disagreement(self, disagreement_probas: torch.Tensor):
        metric = Disagreement()
        metric.update(disagreement_probas)
        res = metric.compute()
        assert res == 1.0

    def test_compute_agreement(self, agreement_probas: torch.Tensor):
        metric = Disagreement()
        metric.update(agreement_probas)
        res = metric.compute()
        assert res == 0.0

    def test_compute_mixed(self, disagreement_probas: torch.Tensor, agreement_probas: torch.Tensor):
        metric = Disagreement()
        metric.update(agreement_probas)
        metric.update(disagreement_probas)
        res = metric.compute()
        assert res == 0.5

    def test_compute_mixed_3_estimators(self, disagreement_probas_3: torch.Tensor):
        metric = Disagreement()
        metric.update(disagreement_probas_3)
        res = metric.compute()
        assert res == pytest.approx(2 / 3, 1e-6)

        metric = Disagreement(reduction="sum")
        metric.update(disagreement_probas_3)
        res = metric.compute()
        assert res == pytest.approx(2 / 3, 1e-6)

        metric = Disagreement(reduction="none")
        metric.update(disagreement_probas_3)
        res = metric.compute()
        assert res == pytest.approx(2 / 3, 1e-6)

    def test_bad_argument_reduction(self):
        with pytest.raises(ValueError):
            _ = Disagreement(reduction="geometric_mean")



# ruff: noqa: F401
import pytest
import torch

from torch_uncertainty.optim_recipes import (
    CosineAnnealingWarmup,
    CosineSWALR,
    get_procedure,
    optim_abnn,
)


class TestCosineAnnealingWarmup:
    def test_full_cosine_annealing_warmup(self):
        CosineAnnealingWarmup(
            torch.optim.SGD(torch.nn.Linear(1, 1).parameters(), lr=1e-3),
            warmup_start_factor=0.1,
            warmup_epochs=5,
            max_epochs=100,
            eta_min=1e-5,
        )


class TestCosineSWALR:
    def test_full_swa_lr(self):
        CosineSWALR(
            torch.optim.SGD(torch.nn.Linear(1, 1).parameters(), lr=1e-3),
            swa_lr=1,
            milestone=12,
            anneal_epochs=5,
        )


class TestOptProcedures:
    def test_optim_cifar10(self):
        model = torch.nn.Linear(1, 1)
        get_procedure("resnet18", "cifar10", "standard")(model)
        get_procedure("resnet34", "cifar10", "masked")(model)
        get_procedure("resnet50", "cifar10", "packed")(model)
        get_procedure("wideresnet28x10", "cifar10", "batched")(model)
        get_procedure("vgg16", "cifar10", "standard")(model)
        optim_abnn(model, lr=0.1)

    def test_optim_cifar100(self):
        model = torch.nn.Linear(1, 1)
        get_procedure("resnet18", "cifar100", "masked")(model)
        get_procedure("resnet34", "cifar100", "masked")(model)
        get_procedure("resnet50", "cifar100")(model)
        get_procedure("wideresnet28x10", "cifar100")(model)
        get_procedure("vgg16", "cifar100", "standard")(model)

    def test_optim_tinyimagenet(self):
        model = torch.nn.Linear(1, 1)
        get_procedure("resnet34", "tiny-imagenet", "standard")(model)
        get_procedure("resnet50", "tiny-imagenet", "standard")(model)

    def test_optim_imagenet_resnet50(self):
        model = torch.nn.Linear(1, 1)
        get_procedure("resnet50", "imagenet", "standard", "A3")(model)
        get_procedure("resnet50", "imagenet", "standard")(model)

    def test_optim_unknown(self):
        with pytest.raises(NotImplementedError):
            _ = get_procedure("unknown", "cifar100")
        with pytest.raises(NotImplementedError):
            _ = get_procedure("resnet18", "unknown")
        with pytest.raises(NotImplementedError):
            _ = get_procedure("resnet34", "unknown")
        with pytest.raises(NotImplementedError):
            _ = get_procedure("resnet50", "unknown")
        with pytest.raises(NotImplementedError):
            _ = get_procedure("wideresnet28x10", "unknown")



from typing import Literal

import torch
import torch.nn.functional as F
from torch import Tensor
from torchmetrics import Metric
from torchmetrics.utilities.data import dim_zero_cat


class BrierScore(Metric):
    is_differentiable = True
    higher_is_better = False
    full_state_update = False

    def __init__(
        self,
        num_classes: int,
        top_class: bool = False,
        reduction: Literal["mean", "sum", "none", None] = "mean",
        **kwargs,
    ) -> None:
        r"""The Brier Score Metric.

        Args:
            num_classes (int): Number of classes
            top_class (bool, optional): If true, compute the Brier score for the
                top class only. Defaults to False.
            reduction (str, optional): Determines how to reduce over the
                :math:`B`/batch dimension:

                - ``'mean'`` [default]: Averages score across samples
                - ``'sum'``: Sum score across samples
                - ``'none'`` or ``None``: Returns score per sample

            kwargs: Additional keyword arguments, see `Advanced metric settings
                <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.

        Inputs:
            - :attr:`probs`: :math:`(B, C)` or :math:`(B, N, C)`
            - :attr:`target`: :math:`(B)` or :math:`(B, C)`

            where :math:`B` is the batch size, :math:`C` is the number of classes
            and :math:`N` is the number of estimators.

        Note:
            If :attr:`probs` is a 3d tensor, then the metric computes the mean of
            the Brier score over the estimators ie. :math:`t = \frac{1}{N}
            \sum_{i=0}^{N-1} BrierScore(probs[:,i,:], target)`.

        Warning:
            Make sure that the probabilities in :attr:`probs` are normalized to sum
            to one.

        Raises:
            ValueError:
                If :attr:`reduction` is not one of ``'mean'``, ``'sum'``,
                ``'none'`` or ``None``.
        """
        super().__init__(**kwargs)

        allowed_reduction = ("sum", "mean", "none", None)
        if reduction not in allowed_reduction:
            raise ValueError(
                "Expected argument `reduction` to be one of ",
                f"{allowed_reduction} but got {reduction}",
            )

        self.num_classes = num_classes
        self.top_class = top_class
        self.reduction = reduction
        self.num_estimators = 1

        if self.reduction in ["mean", "sum"]:
            self.add_state("values", default=torch.tensor(0.0), dist_reduce_fx="sum")
        else:
            self.add_state("values", default=[], dist_reduce_fx="cat")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, probs: Tensor, target: Tensor) -> None:
        """Update the current Brier score with a new tensor of probabilities.

        Args:
            probs (Tensor): A probability tensor of shape
                (batch, num_estimators, num_classes) or
                (batch, num_classes)
            target (Tensor): A tensor of ground truth labels of shape
                (batch, num_classes) or (batch)
        """
        if target.ndim == 1 and self.num_classes > 1:
            target = F.one_hot(target, self.num_classes)

        if probs.ndim <= 2:
            batch_size = probs.size(0)
        elif probs.ndim == 3:
            batch_size = probs.size(0)
            self.num_estimators = probs.size(1)
            target = target.unsqueeze(1).repeat(1, self.num_estimators, 1)
        else:
            raise ValueError(
                f"Expected `probs` to be of shape (batch, num_classes) or "
                f"(batch, num_estimators, num_classes) but got {probs.shape}"
            )

        if self.top_class:
            probs, indices = probs.max(dim=-1)
            target = target.gather(-1, indices.unsqueeze(-1)).squeeze(-1)
            brier_score = F.mse_loss(probs, target, reduction="none")
        else:
            brier_score = F.mse_loss(probs, target, reduction="none").sum(dim=-1)

        if self.reduction is None or self.reduction == "none":
            self.values.append(brier_score)
        else:
            self.values += brier_score.sum()
            self.total += batch_size

    def compute(self) -> Tensor:
        """Compute the final Brier score based on inputs passed to ``update``.

        Returns:
            Tensor: The final value(s) for the Brier score
        """
        values = dim_zero_cat(self.values)
        if self.reduction == "sum":
            return values.sum(dim=-1) / self.num_estimators
        if self.reduction == "mean":
            return values.sum(dim=-1) / self.total / self.num_estimators
        return values



import logging
from typing import Literal

import torch
from torch import Tensor, nn, optim
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm

from torch_uncertainty.post_processing import PostProcessing


class Scaler(PostProcessing):
    criterion = nn.CrossEntropyLoss()
    trained = False

    def __init__(
        self,
        model: nn.Module | None = None,
        lr: float = 0.1,
        max_iter: int = 100,
        device: Literal["cpu", "cuda"] | torch.device | None = None,
    ) -> None:
        """Virtual class for scaling post-processing for calibrated probabilities.

        Args:
            model (nn.Module): Model to calibrate.
            lr (float, optional): Learning rate for the optimizer. Defaults to 0.1.
            max_iter (int, optional): Maximum number of iterations for the
                optimizer. Defaults to 100.
            device (Optional[Literal["cpu", "cuda"]], optional): Device to use
                for optimization. Defaults to None.

        Reference:
            Guo, C., Pleiss, G., Sun, Y., & Weinberger, K. Q. On calibration
            of modern neural networks. In ICML 2017.
        """
        super().__init__(model)
        self.device = device

        if lr <= 0:
            raise ValueError("Learning rate must be positive.")
        self.lr = lr

        if max_iter <= 0:
            raise ValueError("Max iterations must be positive.")
        self.max_iter = int(max_iter)

    def fit(
        self,
        calibration_set: Dataset,
        save_logits: bool = False,
        progress: bool = True,
    ) -> None:
        """Fit the temperature parameters to the calibration data.

        Args:
            calibration_set (Dataset): Calibration dataset.
            save_logits (bool, optional): Whether to save the logits and
                labels. Defaults to False.
            progress (bool, optional): Whether to show a progress bar.
                Defaults to True.
        """
        if self.model is None:
            raise ValueError(
                "Cannot fit a Scaler method without model. Call .set_model(model) first."
            )

        logits_list = []
        labels_list = []
        calibration_dl = DataLoader(calibration_set, batch_size=32, shuffle=False, drop_last=False)
        with torch.no_grad():
            for inputs, labels in tqdm(calibration_dl, disable=not progress):
                logits = self.model(inputs.to(self.device))
                logits_list.append(logits)
                labels_list.append(labels)
        all_logits = torch.cat(logits_list).detach().to(self.device)
        all_labels = torch.cat(labels_list).detach().to(self.device)

        optimizer = optim.LBFGS(self.temperature, lr=self.lr, max_iter=self.max_iter)

        def calib_eval() -> float:
            optimizer.zero_grad()
            loss = self.criterion(self._scale(all_logits), all_labels)
            loss.backward()
            return loss

        optimizer.step(calib_eval)
        self.trained = True
        if save_logits:
            self.logits = logits
            self.labels = labels

    @torch.no_grad()
    def forward(self, inputs: Tensor) -> Tensor:
        if not self.trained:
            logging.error(
                "TemperatureScaler has not been trained yet. Returning manually tempered inputs."
            )
        return self._scale(self.model(inputs))

    def _scale(self, logits: Tensor) -> Tensor:
        """Scale the logits with the optimal temperature.

        Args:
            logits (Tensor): Logits to be scaled.

        Returns:
            Tensor: Scaled logits.
        """
        raise NotImplementedError

    def fit_predict(
        self,
        calibration_set: Dataset,
        progress: bool = True,
    ) -> Tensor:
        self.fit(calibration_set, save_logits=True, progress=progress)
        return self(self.logits)

    @property
    def temperature(self) -> list:
        raise NotImplementedError



from typing import Any

from torch import nn


class Identity(nn.Module):
    # ruff: noqa: ARG002
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__()

    def forward(self, *args) -> Any:
        return args



import matplotlib.pyplot as plt
import torch

from torch_uncertainty.metrics import AUSE


class TestAUSE:
    """Testing the AUSE metric class."""

    def test_compute_zero(self) -> None:
        values = torch.as_tensor([0.1, 0.2, 0.3, 0.4, 0.5])
        metric = AUSE()
        metric.update(values, values)
        assert metric.compute() == 0

    def test_plot(self) -> None:
        scores = torch.as_tensor([0.2, 0.1, 0.5, 0.3, 0.4])
        values = torch.as_tensor([0.1, 0.2, 0.3, 0.4, 0.5])
        metric = AUSE()
        metric.update(scores, values)
        fig, ax = metric.plot()
        assert isinstance(fig, plt.Figure)
        assert isinstance(ax, plt.Axes)
        assert ax.get_xlabel() == "Rejection Rate (%)"
        assert ax.get_ylabel() == "Error Rate (%)"
        plt.close(fig)

        metric = AUSE()
        metric.update(scores, values)
        fig, ax = metric.plot(plot_oracle=False, plot_value=False)
        assert isinstance(fig, plt.Figure)
        assert isinstance(ax, plt.Axes)
        assert ax.get_xlabel() == "Rejection Rate (%)"
        assert ax.get_ylabel() == "Error Rate (%)"
        plt.close(fig)

    def test_compute_nan(self) -> None:
        probs = torch.Tensor([[0.1, 0.9]])
        targets = torch.Tensor([1]).long()
        metric = AUSE()
        assert torch.isnan(metric(probs, targets)).all()



import torch
from torch import nn


class Cutout(nn.Module):
    def __init__(self, length: int, value: int = 0) -> None:
        """Cutout augmentation class.

        Args:
            length (int): Length of the cutout square.
            value (int): Pixel value to be filled in the cutout square.
        """
        super().__init__()

        if length <= 0:
            raise ValueError(f"Cutout length must be positive. Got {length}.")
        self.length = length

        if value < 0 or value > 255:
            raise ValueError(f"Cutout value must be between 0 and 255. Got {value}.")
        self.value = value

    def __call__(self, img: torch.Tensor) -> torch.Tensor:
        if len(img.shape) == 2:
            img = img.unsqueeze(0)
        h, w = img.size(1), img.size(2)
        mask = torch.ones(size=(h, w), dtype=torch.float32)
        y = torch.randint(high=h, size=(1,))
        x = torch.randint(high=w, size=(1,))

        y1 = torch.clip(y - self.length // 2, 0, h)
        y2 = torch.clip(y + self.length // 2, 0, h)
        x1 = torch.clip(x - self.length // 2, 0, w)
        x2 = torch.clip(x + self.length // 2, 0, w)

        mask[y1:y2, x1:x2] = self.value
        return img * mask.expand_as(img)



import json
import logging
import shutil
from collections.abc import Callable
from pathlib import Path
from typing import Literal, NamedTuple

import torch
from einops import rearrange, repeat
from PIL import Image
from torchvision import tv_tensors
from torchvision.datasets import VisionDataset
from torchvision.datasets.utils import (
    download_and_extract_archive,
    download_url,
)
from torchvision.transforms.v2 import functional as F


class CamVidClass(NamedTuple):
    name: str
    index: int
    color: tuple[int, int, int]


class CamVid(VisionDataset):
    # Notes: some classes are not used here
    classes = [
        CamVidClass("animal", 0, (64, 128, 64)),
        CamVidClass("archway", 1, (192, 0, 128)),
        CamVidClass("bicyclist", 2, (0, 128, 192)),
        CamVidClass("bridge", 3, (0, 128, 64)),
        CamVidClass("building", 4, (128, 0, 0)),
        CamVidClass("car", 5, (64, 0, 128)),
        CamVidClass("cart_luggage_pram", 6, (64, 0, 192)),
        CamVidClass("child", 7, (192, 128, 64)),
        CamVidClass("column_pole", 8, (192, 192, 128)),
        CamVidClass("fence", 9, (64, 64, 128)),
        CamVidClass("lane_mkgs_driv", 10, (128, 0, 192)),
        CamVidClass("lane_mkgs_non_driv", 11, (192, 0, 64)),
        CamVidClass("misc_text", 12, (128, 128, 64)),
        CamVidClass("motorcycle_scooter", 13, (192, 0, 192)),
        CamVidClass("othermoving", 14, (128, 64, 64)),
        CamVidClass("parking_block", 15, (64, 192, 128)),
        CamVidClass("pedestrian", 16, (64, 64, 0)),
        CamVidClass("road", 17, (128, 64, 128)),
        CamVidClass("road_shoulder", 18, (128, 128, 192)),
        CamVidClass("sidewalk", 19, (0, 0, 192)),
        CamVidClass("sign_symbol", 20, (192, 128, 128)),
        CamVidClass("sky", 21, (128, 128, 128)),
        CamVidClass("suv_pickup_truck", 22, (64, 128, 192)),
        CamVidClass("traffic_cone", 23, (0, 0, 64)),
        CamVidClass("traffic_light", 24, (0, 64, 64)),
        CamVidClass("train", 25, (192, 64, 128)),
        CamVidClass("tree", 26, (128, 128, 0)),
        CamVidClass("truck_bus", 27, (192, 128, 192)),
        CamVidClass("tunnel", 28, (64, 0, 64)),
        CamVidClass("vegetation_misc", 29, (192, 192, 0)),
        CamVidClass("void", 30, (0, 0, 0)),
        CamVidClass("wall", 31, (64, 192, 0)),
    ]
    superclasses = [
        CamVidClass("sky", 0, (128, 128, 128)),
        CamVidClass("building", 1, (128, 0, 0)),
        CamVidClass("pole", 2, (192, 192, 128)),
        CamVidClass("road", 3, (128, 64, 128)),
        CamVidClass("pavement", 4, (0, 0, 192)),
        CamVidClass("tree", 5, (128, 128, 0)),
        CamVidClass("sign_symbol", 6, (192, 128, 128)),
        CamVidClass("fence", 7, (64, 64, 128)),
        CamVidClass("car", 8, (64, 0, 128)),
        CamVidClass("pedestrian", 9, (64, 64, 0)),
        CamVidClass("bicyclist", 10, (0, 128, 192)),
        CamVidClass("void", None, (0, 0, 0)),
    ]
    superclasses_indices = [
        [21],
        [3, 4, 31, 28, 1],
        [8, 23],
        [17, 10, 11],
        [19, 15, 18],
        [26, 29],
        [20, 12, 24],
        [9],
        [5, 22, 27, 25, 14],
        [16, 7, 6, 0],
        [2, 13],
    ]

    urls = {
        "raw": "http://web4.cs.ucl.ac.uk/staff/g.brostow/MotionSegRecData/files/701_StillsRaw_full.zip",
        "label": "http://web4.cs.ucl.ac.uk/staff/g.brostow/MotionSegRecData/data/LabeledApproved_full.zip",
        "splits": "https://raw.githubusercontent.com/torch-uncertainty/dataset-metadata/main/segmentation/camvid/splits.json",
    }

    splits_md5 = "db45289aaa83c60201391b11e78c6382"

    filenames = {
        "raw": "701_StillsRaw_full.zip",
        "label": "LabeledApproved_full.zip",
    }
    base_folder = "camvid"
    num_samples = {
        "train": 367,
        "val": 101,
        "test": 233,
        "all": 701,
    }

    def __init__(
        self,
        root: str,
        group_classes: bool = True,
        split: Literal["train", "val", "test"] | None = None,
        transforms: Callable | None = None,
        download: bool = False,
    ) -> None:
        """`CamVid <http://web4.cs.ucl.ac.uk/staff/g.brostow/MotionSegRecData/>`_ Dataset.

        Args:
            root (str): Root directory of dataset where ``camvid/`` exists or
                will be saved to if download is set to ``True``.
            group_classes (bool, optional): Whether to group the 32 classes into
                11 superclasses. Default: ``True``.
            split (str, optional): The dataset split, supports ``train``,
                ``val`` and ``test``. Default: ``None``.
            transforms (callable, optional): A function/transform that takes
                input sample and its target as entry and returns a transformed
                version. Default: ``None``.
            download (bool, optional): If true, downloads the dataset from the
                internet and puts it in root directory. If dataset is already
                downloaded, it is not downloaded again.
        """
        if split not in ["train", "val", "test", None]:
            raise ValueError(
                f"Unknown split '{split}'. Supported splits are ['train', 'val', 'test', None]"
            )

        super().__init__(root, transforms, None, None)
        self.group_classes = group_classes
        self.class_to_superclass = []
        for i in range(32):
            if i == 30:  # For void
                self.class_to_superclass.append(None)

            for j, superclass in enumerate(self.superclasses_indices):
                if i in superclass:
                    self.class_to_superclass.append(j)
                    break

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it"
            )

        # get filenames for split
        if split is None:
            self.images = sorted((Path(self.root) / "camvid" / "raw").glob("*.png"))
            self.targets = sorted((Path(self.root) / "camvid" / "label").glob("*.png"))
        else:
            with (Path(self.root) / "camvid" / "splits.json").open() as f:
                filenames = json.load(f)[split]

            self.images = sorted(
                [
                    path
                    for path in (Path(self.root) / "camvid" / "raw").glob("*.png")
                    if path.stem in filenames
                ]
            )
            self.targets = sorted(
                [
                    path
                    for path in (Path(self.root) / "camvid" / "label").glob("*.png")
                    if path.stem[:-2] in filenames
                ]
            )

        self.split = split if split is not None else "all"

    def encode_target(self, target: Image.Image) -> torch.Tensor:
        """Encode target image to tensor.

        Args:
            target (Image.Image): Target PIL image.

        Returns:
            torch.Tensor: Encoded target.
        """
        colored_target = F.pil_to_tensor(target)
        colored_target = rearrange(colored_target, "c h w -> h w c")
        target = torch.zeros_like(colored_target[..., :1])

        # convert target color to index
        for camvid_class in self.classes:
            index = camvid_class.index if camvid_class.index != 30 else 255
            if self.group_classes and index != 255:
                index = self.class_to_superclass[index]
            target[
                (colored_target == torch.tensor(camvid_class.color, dtype=target.dtype)).all(dim=-1)
            ] = index

        return rearrange(target, "h w c -> c h w")

    def decode_target(self, target: torch.Tensor) -> Image.Image:
        """Decode target tensor to image.

        Args:
            target (torch.Tensor): Target tensor.

        Returns:
            Image.Image: Decoded target as a PIL.Image.
        """
        colored_target = repeat(target.clone(), "h w  -> h w 3", c=3)
        if not self.group_classes:
            for camvid_class in self.classes:
                colored_target[
                    (target == torch.tensor(camvid_class.index, dtype=target.dtype)).all(dim=0)
                ] = torch.tensor(camvid_class.color, dtype=target.dtype)
        else:
            for camvid_class in self.superclasses:
                colored_target[
                    (target == torch.tensor(camvid_class.index, dtype=target.dtype)).all(dim=0)
                ] = torch.tensor(camvid_class.color, dtype=target.dtype)
        return F.to_pil_image(rearrange(colored_target, "h w c -> c h w"))

    @property
    def color_palette(self) -> list[tuple[int, int, int]]:
        """Return the color palette of the dataset."""
        if self.group_classes:
            return [camvid_class.color for camvid_class in self.superclasses]
        return [camvid_class.color for camvid_class in self.classes]

    def __getitem__(self, index: int) -> tuple[tv_tensors.Image, tv_tensors.Mask]:
        """Get the image and target at the given index.

        Args:
            index (int): Sample index.

        Returns:
            tuple[tv_tensors.Image, tv_tensors.Mask]: Image and target.
        """
        image = tv_tensors.Image(Image.open(self.images[index]).convert("RGB"))
        target = tv_tensors.Mask(self.encode_target(Image.open(self.targets[index])))

        if self.transforms is not None:
            image, target = self.transforms(image, target)

        return image, target

    def __len__(self) -> int:
        """Return the number of samples."""
        return self.num_samples[self.split]

    def _check_integrity(self) -> bool:
        """Check if the dataset exists."""
        if len(list((Path(self.root) / "camvid" / "raw").glob("*.png"))) != self.num_samples["all"]:
            return False
        if (
            len(list((Path(self.root) / "camvid" / "label").glob("*.png")))
            != self.num_samples["all"]
        ):
            return False

        return (Path(self.root) / "camvid" / "splits.json").exists()

    def download(self) -> None:
        """Download the CamVid data if it doesn't exist already."""
        if self._check_integrity():
            logging.info("Files already downloaded and verified")
            return

        if (Path(self.root) / self.base_folder).exists():
            shutil.rmtree(Path(self.root) / self.base_folder)

        download_and_extract_archive(
            self.urls["raw"],
            self.root,
            extract_root=Path(self.root) / "camvid",
            filename=self.filenames["raw"],
        )
        (Path(self.root) / "camvid" / "701_StillsRaw_full").replace(
            Path(self.root) / "camvid" / "raw"
        )
        download_and_extract_archive(
            self.urls["label"],
            self.root,
            extract_root=Path(self.root) / "camvid" / "label",
            filename=self.filenames["label"],
        )
        download_url(
            self.urls["splits"],
            Path(self.root) / "camvid",
            filename="splits.json",
            md5=self.splits_md5,
        )



from importlib import util

import torch

if util.find_spec("glest"):
    from glest import GLEstimator as GLEstimatorBase

    glest_installed = True
else:  # coverage: ignore
    glest_installed = False
    GLEstimatorBase = object

from torch import Tensor
from torchmetrics import Metric
from torchmetrics.utilities import rank_zero_warn
from torchmetrics.utilities.data import dim_zero_cat


class GLEstimator(GLEstimatorBase):
    def fit(self, probs: Tensor, targets: Tensor, features: Tensor) -> "GLEstimator":
        probs = probs.detach().cpu().numpy()
        features = features.detach().cpu().numpy()
        targets = (targets * 1).detach().cpu().numpy()
        self.classifier = probs
        return super().fit(features, targets)


class GroupingLoss(Metric):
    is_differentiable = False
    higher_is_better = False
    full_state_update = False

    def __init__(
        self,
        **kwargs,
    ) -> None:
        r"""Metric to estimate the Top-label Grouping Loss.

        Args:
            kwargs: Additional keyword arguments, see `Advanced metric settings
                <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.

        Inputs:
            - :attr:`probs`: :math:`(B, C)` or :math:`(B, N, C)`
            - :attr:`target`: :math:`(B)` or :math:`(B, C)`
            - :attr:`features`: :math:`(B, F)` or :math:`(B, N, F)`

            where :math:`B` is the batch size, :math:`C` is the number of classes
            and :math:`N` is the number of estimators.

        Warning:
            Make sure that the probabilities in :attr:`probs` are normalized to sum
            to one.

        Raises:
            ValueError:
                If :attr:`reduction` is not one of ``'mean'``, ``'sum'``,
                ``'none'`` or ``None``.

        Reference:
            Perez-Lebel, Alexandre, Le Morvan, Marine and Varoquaux, Gaël.
            Beyond calibration: estimating the grouping loss of modern neural
            networks. In ICLR 2023.
        """
        super().__init__(**kwargs)
        if not glest_installed:  # coverage: ignore
            raise ImportError(
                "The glest library is not installed. Please install"
                "torch_uncertainty with the all option:"
                """pip install -U "torch_uncertainty[all]"."""
            )

        self.estimator = GLEstimator(None)

        self.add_state("probs", default=[], dist_reduce_fx="cat")
        self.add_state("targets", default=[], dist_reduce_fx="cat")
        self.add_state("features", default=[], dist_reduce_fx="cat")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

        rank_zero_warn(
            "Metric `GroupingLoss` will save all targets, predictions and features"
            " in buffer. For large datasets this may lead to large memory"
            " footprint."
        )

    def update(self, probs: Tensor, target: Tensor, features: Tensor) -> None:
        """Accumulate the tensors for the estimation of the Grouping Loss.

        Args:
            probs (Tensor): A probability tensor of shape (batch, num_classes),
                (batch, num_estimators, num_classes), or (batch) if binary
                classification
            target (Tensor): A tensor of ground truth labels of shape
                (batch, num_classes) or (batch)
            features (Tensor): A tensor of features of shape
                (batch, num_estimators, num_features) or (batch, num_features)
        """
        if target.ndim == 2:
            target = target.argmax(dim=-1)
        elif target.ndim != 1:
            raise ValueError(
                "Expected `target` to be of shape (batch) or (batch, num_classes) "
                f"but got {target.shape}."
            )

        if probs.ndim == 1:
            self.probs.append(probs)
            self.targets.append(target == (probs > 0.5).int())
        elif probs.ndim == 2:
            max_probs = probs.max(-1)
            self.probs.append(max_probs.values)
            self.targets.append(target == max_probs.indices)
        elif probs.ndim == 3:
            max_probs = probs.mean(1).max(-1)
            self.probs.append(max_probs.values)
            self.targets.append(target == max_probs.indices)
        else:
            raise ValueError(
                "Expected `probs` to be of shape (batch, num_classes) or "
                "(batch, num_estimators, num_classes) or (batch) "
                f"but got {probs.shape}."
            )

        if features.ndim == 2:
            self.features.append(features)
        elif features.ndim == 3:
            self.features.append(features[:, 0, :])
        else:
            raise ValueError(
                "Expected `features` to be of shape (batch, num_features) or "
                "(batch, num_estimators, num_features) but got "
                f"{features.shape}."
            )

    def compute(self) -> Tensor:
        """Compute the final Brier score based on inputs passed to ``update``.

        Returns:
            torch.Tensor: The final value(s) for the Brier score
        """
        probs = dim_zero_cat(self.probs)
        features = dim_zero_cat(self.features)
        targets = dim_zero_cat(self.targets)
        estimator = self.estimator.fit(probs, targets, features)
        return estimator.metrics("brier")["GL"]



from typing import Any

import torch
from torch import Tensor
from torchmetrics import Metric
from torchmetrics.utilities.data import dim_zero_cat


class SILog(Metric):
    def __init__(self, sqrt: bool = False, lmbda: float = 1.0, **kwargs: Any) -> None:
        r"""The Scale-Invariant Logarithmic Loss metric.

        .. math:: \text{SILog} = \frac{1}{N} \sum_{i=1}^{N} \left(\log(y_i) - \log(\hat{y_i})\right)^2 - \left(\frac{1}{N} \sum_{i=1}^{N} \log(y_i) \right)^2,

        where :math:`N` is the batch size, :math:`y_i` is a tensor of target
        values and :math:`\hat{y_i}` is a tensor of prediction.
        Return the square root of SILog by setting :attr:`sqrt` to `True`.

        Args:
            sqrt: If `True`, return the square root of the metric. Defaults to
                False.
            lmbda: The regularization parameter on the variance of error.
                Defaults to 1.0.
            kwargs: Additional keyword arguments, see `Advanced metric settings
                <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.

        Reference:
            Depth Map Prediction from a Single Image using a Multi-Scale Deep Network. David Eigen, Christian Puhrsch, Rob Fergus. NeurIPS 2014.
            From Big to Small: Multi-Scale Local Planar Guidance for Monocular Depth Estimation. Jin Han Lee, Myung-Kyu Han, Dong Wook Ko and Il Hong Suh. (For :attr:`lmbda`)
        """
        super().__init__(**kwargs)
        self.sqrt = sqrt
        self.lmbda = lmbda
        self.add_state(
            "log_dists",
            default=torch.tensor(0.0),
            dist_reduce_fx="sum",
        )
        self.add_state(
            "sq_log_dists",
            default=torch.tensor(0.0),
            dist_reduce_fx="sum",
        )
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, pred: Tensor, target: Tensor) -> None:
        """Update state with predictions and targets.

        Args:
            pred (Tensor): A prediction tensor of shape (batch)
            target (Tensor): A tensor of ground truth labels of shape (batch)
        """
        self.log_dists += torch.sum(pred.log() - target.log())
        self.sq_log_dists += torch.sum((pred.log() - target.log()) ** 2)
        self.total += target.size(0)

    def compute(self) -> Tensor:
        """Compute the Scale-Invariant Logarithmic Loss."""
        log_dists = dim_zero_cat(self.log_dists)
        sq_log_dists = dim_zero_cat(self.sq_log_dists)
        out = sq_log_dists / self.total - self.lmbda * log_dists**2 / (self.total * self.total)
        if self.sqrt:
            return torch.sqrt(out)
        return out



from typing import Literal

import torch
from torch import nn

from .scaler import Scaler


class VectorScaler(Scaler):
    def __init__(
        self,
        num_classes: int,
        model: nn.Module | None = None,
        init_w: float = 1,
        init_b: float = 0,
        lr: float = 0.1,
        max_iter: int = 200,
        device: Literal["cpu", "cuda"] | torch.device | None = None,
    ) -> None:
        """Vector scaling post-processing for calibrated probabilities.

        Args:
            model (nn.Module): Model to calibrate.
            num_classes (int): Number of classes.
            init_w (float, optional): Initial value for the weights.
                Defaults to 1.
            init_b (float, optional): Initial value for the bias.
                Defaults to 0.
            lr (float, optional): Learning rate for the optimizer. Defaults to 0.1.
            max_iter (int, optional): Maximum number of iterations for the
                optimizer. Defaults to 100.
            device (Optional[Literal["cpu", "cuda"]], optional): Device to use
                for optimization. Defaults to None.

        Reference:
            Guo, C., Pleiss, G., Sun, Y., & Weinberger, K. Q. On calibration
            of modern neural networks. In ICML 2017.

        """
        super().__init__(model=model, lr=lr, max_iter=max_iter, device=device)

        if not isinstance(num_classes, int):
            raise TypeError("num_classes must be an integer.")
        if num_classes <= 0:
            raise ValueError("The number of classes must be positive.")
        self.num_classes = num_classes

        self.set_temperature(init_w, init_b)

    def set_temperature(self, val_w: float, val_b: float) -> None:
        """Set the temperature to a fixed value.

        Args:
            val_w (float): Weight temperature value.
            val_b (float): Bias temperature value.
        """
        self.temp_w = nn.Parameter(
            torch.ones(self.num_classes, device=self.device) * val_w,
            requires_grad=True,
        )
        self.temp_b = nn.Parameter(
            torch.ones(self.num_classes, device=self.device) * val_b,
            requires_grad=True,
        )

    def _scale(self, logits: torch.Tensor) -> torch.Tensor:
        """Scale the predictions with the optimal temperature.

        Args:
            logits (torch.Tensor): logits to be scaled.

        Returns:
            torch.Tensor: Scaled logits.
        """
        return self.temp_w * logits + self.temp_b

    @property
    def temperature(self) -> list:
        return [self.temp_w, self.temp_b]



from functools import partial

from tests._dummies.dataset import DummyRegressionDataset
from torch_uncertainty.datamodules import UCIRegressionDataModule


class TestUCIRegressionDataModule:
    """Testing the UCIRegressionDataModule datamodule class."""

    def test_uci_regression(self):
        dm = UCIRegressionDataModule(dataset_name="kin8nm", root="./data/", batch_size=128)

        dm.dataset = partial(DummyRegressionDataset, num_samples=64)
        dm.prepare_data()
        dm.val_split = 0.5
        dm.setup()

        dm.train_dataloader()
        dm.val_dataloader()
        dm.test_dataloader()

        dm.val_split = 0
        dm.setup()



from collections.abc import Callable
from pathlib import Path
from typing import Any

import numpy as np
import torch
from PIL import Image
from torch.utils.data import Dataset
from torchvision import tv_tensors


class DummyClassificationDataset(Dataset):
    """Dummy dataset for testing purposes.

    Args:
        root (string): Root directory containing the dataset (unused).
        train (bool, optional): If True, creates dataset from training set,
            otherwise creates from test set (unused).
        transform (callable, optional): A function/transform that takes in
            a PIL image and returns a transformed version. E.g,
            ``transforms.RandomCrop``
        target_transform (callable, optional): A function/transform that
            takes in the target and transforms it.
        num_channels (int, optional): Number of channels in the images.
        image_size (int, optional): Size of the images.
        num_classes (int, optional): Number of classes in the dataset.
        num_images (int, optional): Number of images in the dataset.
        kwargs (Any): Other arguments.
    """

    def __init__(
        self,
        root: Path,
        train: bool = True,
        transform: Callable[..., Any] | None = None,
        target_transform: Callable[..., Any] | None = None,
        num_channels: int = 3,
        image_size: int = 4,
        num_classes: int = 10,
        num_images: int = 2,
        **args,
    ) -> None:
        self.root = root
        self.train = train  # training set or test set
        self.transform = transform
        self.target_transform = target_transform

        self.data: Any = []
        self.targets = []

        if num_channels == 1:
            shape = (num_images, image_size, image_size)
        else:
            shape = (num_images, num_channels, image_size, image_size)

        self.data = np.random.default_rng().integers(
            low=0,
            high=255,
            size=shape,
            dtype=np.uint8,
        )

        if num_channels == 1:
            self.data = self.data.transpose((0, 1, 2))  # convert to HWC
        else:
            self.data = self.data.transpose((0, 2, 3, 1))  # convert to HWC

        self.targets = torch.randint(low=0, high=num_classes, size=(num_images,))
        self.targets = torch.arange(start=0, end=num_classes).repeat(
            num_images // (num_classes) + 1
        )[:num_images]

        self.samples = self.data  # for compatibility with TinyImagenet
        self.label_data = self.targets

    def __getitem__(self, index: int) -> tuple[Any, Any]:
        """Get item from dataset.

        Args:
            index (int): Index.

        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        img, target = self.data[index], self.targets[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self) -> int:
        return len(self.data)


class DummyRegressionDataset(Dataset):
    def __init__(
        self,
        root: str,
        train: bool = True,
        transform: Callable[..., Any] | None = None,
        target_transform: Callable[..., Any] | None = None,
        in_features: int = 3,
        out_features: int = 10,
        num_samples: int = 2,
        **args,
    ) -> None:
        self.root = root
        self.train = train  # training set or test set
        self.transform = transform
        self.target_transform = target_transform

        self.data: Any = []
        self.targets = []

        input_shape = (num_samples, in_features)

        output_shape = (num_samples, out_features) if out_features != 1 else (num_samples,)

        self.data = torch.rand(
            size=input_shape,
        )

        self.targets = torch.rand(
            size=output_shape,
        )

    def __getitem__(self, index: int) -> tuple[Any, Any]:
        """Get item from dataset.

        Args:
            index (int): Index.

        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        img, target = self.data[index], self.targets[index]

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self) -> int:
        return len(self.data)


class DummySegmentationDataset(Dataset):
    def __init__(
        self,
        root: Path,
        split: str = "train",
        transforms: Callable[..., Any] | None = None,
        num_channels: int = 3,
        image_size: int = 4,
        num_classes: int = 10,
        num_images: int = 2,
        **args,
    ) -> None:
        super().__init__()

        self.root = root
        self.split = split
        self.transforms = transforms

        self.data: Any = []
        self.targets = []

        if num_channels == 1:
            img_shape = (num_images, image_size, image_size)
        else:
            img_shape = (num_images, num_channels, image_size, image_size)

        smnt_shape = (num_images, 1, image_size, image_size)

        rng = np.random.default_rng()
        self.data = rng.integers(
            low=0,
            high=255,
            size=img_shape,
            dtype=np.uint8,
        )

        self.targets = rng.integers(
            low=0,
            high=num_classes,
            size=smnt_shape,
            dtype=np.uint8,
        )

    def __getitem__(self, index: int) -> tuple[Any, Any]:
        img = tv_tensors.Image(self.data[index])
        target = tv_tensors.Mask(self.targets[index])

        if self.transforms is not None:
            img, target = self.transforms(img, target)

        return img, target

    def __len__(self) -> int:
        return len(self.data)


class DummPixelRegressionDataset(Dataset):
    def __init__(
        self,
        root: Path,
        split: str = "train",
        transforms: Callable[..., Any] | None = None,
        input_channels: int = 3,
        image_size: int = 4,
        num_images: int = 2,
        output_dim: int = 1,
        **args,
    ) -> None:
        super().__init__()

        self.root = root
        self.split = split
        self.transforms = transforms

        self.data: Any = []
        self.targets = []

        if input_channels == 1:
            img_shape = (num_images, image_size, image_size)
        else:
            img_shape = (num_images, input_channels, image_size, image_size)

        if output_dim == 1:
            smnt_shape = (num_images, image_size, image_size)
        else:
            smnt_shape = (num_images, output_dim, image_size, image_size)

        rng = np.random.default_rng()
        self.data = rng.integers(
            low=0,
            high=255,
            size=img_shape,
            dtype=np.uint8,
        )

        self.targets = rng.uniform(
            low=0,
            high=100,
            size=smnt_shape,
        )

    def __getitem__(self, index: int) -> tuple[Any, Any]:
        img = tv_tensors.Image(self.data[index])
        target = tv_tensors.Mask(self.targets[index])

        if self.transforms is not None:
            img, target = self.transforms(img, target)

        return img, target

    def __len__(self) -> int:
        return len(self.data)



import pytest
import torch
from torch import nn, softmax

from torch_uncertainty.post_processing import (
    MatrixScaler,
    TemperatureScaler,
    VectorScaler,
)


class TestTemperatureScaler:
    """Testing the TemperatureScaler class."""

    def test_main(self):
        scaler = TemperatureScaler(model=nn.Identity(), init_val=2)
        scaler.set_temperature(1)

        logits = torch.tensor([[1, 2, 3]], dtype=torch.float32)

        assert scaler.temperature[0].item() == 1.0
        assert torch.all(scaler(logits) == logits)

    def test_fit_biased(self):
        inputs = torch.as_tensor([0.6, 0.4]).repeat(10, 1)
        labels = torch.as_tensor([0.5, 0.5]).repeat(10, 1)

        calibration_set = list(zip(inputs, labels, strict=True))

        scaler = TemperatureScaler(model=nn.Identity(), init_val=2, lr=1, max_iter=10)
        assert scaler.temperature[0] == 2.0
        scaler.fit(calibration_set)
        assert scaler.temperature[0] > 10  # best is +inf
        assert (
            torch.sum(
                softmax(scaler(torch.as_tensor([[0.6, 0.4]])).detach(), dim=1)
                - torch.as_tensor([[0.5, 0.5]])
            )
            ** 2
            < 0.001
        )
        scaler.fit_predict(calibration_set, progress=False)

    def test_errors(self):
        with pytest.raises(ValueError):
            TemperatureScaler(model=nn.Identity(), init_val=-1)

        with pytest.raises(ValueError):
            TemperatureScaler(model=nn.Identity(), lr=-1)

        with pytest.raises(ValueError):
            TemperatureScaler(model=nn.Identity(), max_iter=-1)

        scaler = TemperatureScaler(
            model=nn.Identity(),
        )
        with pytest.raises(ValueError):
            scaler.set_temperature(val=-1)

        scaler = TemperatureScaler(
            model=None,
        )
        with pytest.raises(ValueError, match="Cannot fit a Scaler method without model."):
            scaler.fit(None)


class TestVectorScaler:
    """Testing the VectorScaler class."""

    def test_main(self):
        scaler = VectorScaler(model=nn.Identity(), num_classes=1, init_w=2)
        scaler.set_temperature(1, 0)

        logits = torch.tensor([[1, 2, 3]], dtype=torch.float32)

        assert scaler.temp_w.item() == 1.0
        assert scaler.temp_b.item() == 0.0
        assert torch.all(scaler(logits) == logits)

        _ = scaler.temperature

    def test_errors(self):
        with pytest.raises(ValueError):
            VectorScaler(model=nn.Identity(), num_classes=-1)

        with pytest.raises(TypeError):
            VectorScaler(model=nn.Identity(), num_classes=1.8)

        with pytest.raises(ValueError):
            VectorScaler(model=nn.Identity(), num_classes=2, lr=-1)

        with pytest.raises(ValueError):
            VectorScaler(model=nn.Identity(), num_classes=2, max_iter=-1)


class TestMatrixScaler:
    """Testing the MatrixScaler class."""

    def test_main(self):
        scaler = MatrixScaler(model=nn.Identity(), num_classes=1, init_w=2)
        scaler.set_temperature(1, 0)

        logits = torch.tensor([[1, 2, 3]], dtype=torch.float32)

        assert scaler.temp_w.item() == 1.0
        assert scaler.temp_b.item() == 0.0
        assert torch.all(scaler(logits) == logits)

        _ = scaler.temperature

    def test_errors(self):
        with pytest.raises(ValueError):
            MatrixScaler(model=nn.Identity(), num_classes=-1)

        with pytest.raises(TypeError):
            MatrixScaler(model=nn.Identity(), num_classes=1.8)

        with pytest.raises(ValueError):
            MatrixScaler(model=nn.Identity(), num_classes=2, lr=-1)

        with pytest.raises(ValueError):
            MatrixScaler(model=nn.Identity(), num_classes=2, max_iter=-1)



from collections.abc import Callable
from typing import Literal

from torch import Tensor, nn, relu

from torch_uncertainty.layers.bayesian.lpbnn import LPBNNConv2d, LPBNNLinear

from .utils import get_resnet_num_blocks

__all__ = [
    "lpbnn_resnet",
]


class _BasicBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        in_planes: int,
        planes: int,
        stride: int,
        dropout_rate: float,
        num_estimators: int,
        groups: int,
        activation_fn: Callable,
        normalization_layer: type[nn.Module],
        conv_bias: bool,
    ) -> None:
        super().__init__()
        self.activation_fn = activation_fn

        self.conv1 = LPBNNConv2d(
            in_planes,
            planes,
            kernel_size=3,
            num_estimators=num_estimators,
            groups=groups,
            stride=stride,
            padding=1,
            bias=conv_bias,
        )
        self.bn1 = normalization_layer(planes)
        self.dropout = nn.Dropout2d(p=dropout_rate)
        self.conv2 = LPBNNConv2d(
            planes,
            planes,
            kernel_size=3,
            num_estimators=num_estimators,
            groups=groups,
            stride=1,
            padding=1,
            bias=conv_bias,
        )
        self.bn2 = normalization_layer(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_planes,
                    self.expansion * planes,
                    groups=groups,
                    kernel_size=1,
                    stride=stride,
                    bias=conv_bias,
                ),
                normalization_layer(self.expansion * planes),
            )

    def forward(self, inputs: Tensor) -> Tensor:
        out = self.activation_fn(self.dropout(self.bn1(self.conv1(inputs))))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(inputs)
        return self.activation_fn(out)


class _Bottleneck(nn.Module):
    expansion = 4

    def __init__(
        self,
        in_planes: int,
        planes: int,
        stride: int,
        num_estimators: int,
        dropout_rate: float,
        groups: int,
        activation_fn: Callable,
        normalization_layer: type[nn.Module],
        conv_bias: bool,
    ) -> None:
        super().__init__()
        self.activation_fn = activation_fn

        self.conv1 = LPBNNConv2d(
            in_planes,
            planes,
            kernel_size=1,
            num_estimators=num_estimators,
            groups=groups,
            bias=conv_bias,
        )
        self.bn1 = normalization_layer(planes)
        self.conv2 = LPBNNConv2d(
            planes,
            planes,
            kernel_size=3,
            num_estimators=num_estimators,
            groups=groups,
            stride=stride,
            padding=1,
            bias=conv_bias,
        )
        self.bn2 = normalization_layer(planes)
        self.dropout = nn.Dropout2d(p=dropout_rate)
        self.conv3 = LPBNNConv2d(
            planes,
            self.expansion * planes,
            num_estimators=num_estimators,
            groups=groups,
            kernel_size=1,
            bias=conv_bias,
        )
        self.bn3 = normalization_layer(self.expansion * planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                LPBNNConv2d(
                    in_planes,
                    self.expansion * planes,
                    kernel_size=1,
                    num_estimators=num_estimators,
                    groups=groups,
                    stride=stride,
                    bias=conv_bias,
                ),
                normalization_layer(self.expansion * planes),
            )

    def forward(self, x: Tensor) -> Tensor:
        out = self.activation_fn(self.bn1(self.conv1(x)))
        out = self.activation_fn(self.dropout(self.bn2(self.conv2(out))))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        return self.activation_fn(out)


class _LPBNNResNet(nn.Module):
    def __init__(
        self,
        block: type[_BasicBlock | _Bottleneck],
        num_blocks: list[int],
        in_channels: int,
        num_estimators: int,
        num_classes: int,
        conv_bias: bool,
        dropout_rate: float,
        groups: int,
        style: Literal["imagenet", "cifar"] = "imagenet",
        in_planes: int = 64,
        activation_fn: Callable = relu,
        normalization_layer: type[nn.Module] = nn.BatchNorm2d,
    ):
        super().__init__()
        self.in_planes = in_planes
        block_planes = in_planes
        self.dropout_rate = dropout_rate
        self.activation_fn = activation_fn
        self.num_estimators = num_estimators

        if style == "imagenet":
            self.conv1 = LPBNNConv2d(
                in_channels,
                block_planes,
                kernel_size=7,
                stride=2,
                padding=3,
                num_estimators=num_estimators,
                groups=groups,
                bias=conv_bias,
            )
        elif style == "cifar":
            self.conv1 = LPBNNConv2d(
                in_channels,
                block_planes,
                kernel_size=3,
                stride=1,
                padding=1,
                num_estimators=num_estimators,
                groups=groups,
                bias=conv_bias,
            )
        else:
            raise ValueError(f"Unknown style. Got {style}.")

        self.bn1 = normalization_layer(block_planes)

        if style == "imagenet":
            self.optional_pool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        else:
            self.optional_pool = nn.Identity()

        self.layer1 = self._make_layer(
            block,
            block_planes,
            num_blocks[0],
            stride=1,
            dropout_rate=dropout_rate,
            groups=groups,
            activation_fn=activation_fn,
            normalization_layer=normalization_layer,
            conv_bias=conv_bias,
            num_estimators=num_estimators,
        )
        self.layer2 = self._make_layer(
            block,
            block_planes * 2,
            num_blocks[1],
            stride=2,
            dropout_rate=dropout_rate,
            groups=groups,
            activation_fn=activation_fn,
            normalization_layer=normalization_layer,
            conv_bias=conv_bias,
            num_estimators=num_estimators,
        )
        self.layer3 = self._make_layer(
            block,
            block_planes * 4,
            num_blocks[2],
            stride=2,
            dropout_rate=dropout_rate,
            groups=groups,
            activation_fn=activation_fn,
            normalization_layer=normalization_layer,
            conv_bias=conv_bias,
            num_estimators=num_estimators,
        )
        if len(num_blocks) == 4:
            self.layer4 = self._make_layer(
                block,
                block_planes * 8,
                num_blocks[3],
                stride=2,
                dropout_rate=dropout_rate,
                groups=groups,
                activation_fn=activation_fn,
                normalization_layer=normalization_layer,
                conv_bias=conv_bias,
                num_estimators=num_estimators,
            )
            linear_multiplier = 8
        else:
            self.layer4 = nn.Identity()
            linear_multiplier = 4

        self.final_dropout = nn.Dropout(p=dropout_rate)
        self.pool = nn.AdaptiveAvgPool2d(output_size=1)
        self.flatten = nn.Flatten(1)

        self.linear = LPBNNLinear(
            block_planes * linear_multiplier * block.expansion,
            num_classes,
            num_estimators=num_estimators,
        )

    def _make_layer(
        self,
        block: type[_BasicBlock | _Bottleneck],
        planes: int,
        num_blocks: int,
        stride: int,
        num_estimators: int,
        dropout_rate: float,
        groups: int,
        activation_fn: Callable,
        normalization_layer: type[nn.Module],
        conv_bias: bool,
    ):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(
                block(
                    in_planes=self.in_planes,
                    planes=planes,
                    stride=stride,
                    dropout_rate=dropout_rate,
                    groups=groups,
                    activation_fn=activation_fn,
                    normalization_layer=normalization_layer,
                    conv_bias=conv_bias,
                    num_estimators=num_estimators,
                )
            )
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def feats_forward(self, x: Tensor) -> Tensor:
        out = x.repeat(self.num_estimators, 1, 1, 1)
        out = self.activation_fn(self.bn1(self.conv1(out)))
        out = self.optional_pool(out)
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = self.pool(out)
        return self.final_dropout(self.flatten(out))

    def forward(self, x: Tensor) -> Tensor:
        return self.linear(self.feats_forward(x))


def lpbnn_resnet(
    in_channels: int,
    num_classes: int,
    arch: int,
    num_estimators: int,
    dropout_rate: float = 0,
    conv_bias: bool = True,
    width_multiplier: float = 1.0,
    groups: int = 1,
    style: Literal["imagenet", "cifar"] = "imagenet",
) -> _LPBNNResNet:
    block = _BasicBlock if arch in [18, 20, 34, 44, 56, 110, 1202] else _Bottleneck
    in_planes = 16 if arch in [20, 44, 56, 110, 1202] else 64
    return _LPBNNResNet(
        block=block,
        num_blocks=get_resnet_num_blocks(arch),
        in_channels=in_channels,
        num_estimators=num_estimators,
        num_classes=num_classes,
        dropout_rate=dropout_rate,
        conv_bias=conv_bias,
        groups=groups,
        style=style,
        in_planes=int(in_planes * width_multiplier),
    )



from typing import Any

import torch
import torchvision.transforms.v2.functional as F
from PIL import Image, ImageEnhance
from torch import Tensor, nn
from torchvision.transforms.v2 import InterpolationMode, Transform
from torchvision.transforms.v2._utils import query_size


class AutoContrast(nn.Module):
    pixmix_max_level = None
    level_type = None
    corruption_overlap = False

    def forward(self, img: Tensor | Image.Image) -> Tensor | Image.Image:
        return F.autocontrast(img)


class Equalize(nn.Module):
    pixmix_max_level = None
    level_type = None
    corruption_overlap = False

    def forward(self, img: Tensor | Image.Image) -> Tensor | Image.Image:
        return F.equalize(img)


class Posterize(nn.Module):
    max_level = 4
    pixmix_max_level = 4
    level_type = int
    corruption_overlap = False

    def forward(self, img: Tensor | Image.Image, level: int) -> Tensor | Image.Image:
        if level >= self.max_level:
            raise ValueError(f"Level must be less than {self.max_level}.")
        if level < 0:
            raise ValueError("Level must be greater than 0.")
        return F.posterize(img, self.max_level - level)


class Solarize(nn.Module):
    max_level = 256
    pixmix_max_level = 256
    level_type = int
    corruption_overlap = False

    def forward(self, img: Tensor | Image.Image, level: float) -> Tensor | Image.Image:
        if level >= self.max_level:
            raise ValueError(f"Level must be less than {self.max_level}.")
        if level < 0:
            raise ValueError("Level must be greater than 0.")
        return F.solarize(img, self.max_level - level)


class Rotate(nn.Module):
    pixmix_max_level = 30
    level_type = float
    corruption_overlap = False

    def __init__(
        self,
        random_direction: bool = True,
        interpolation: F.InterpolationMode = F.InterpolationMode.NEAREST,
        expand: bool = False,
        center: list[float] | None = None,
        fill: list[float] | None = None,
    ) -> None:
        super().__init__()
        self.random_direction = random_direction
        self.interpolation = interpolation
        self.expand = expand
        self.center = center
        self.fill = fill

    def forward(self, img: Tensor | Image.Image, level: float) -> Tensor | Image.Image:
        if self.random_direction and torch.rand(1).item() > 0.5:  # coverage: ignore
            level = -level
        return F.rotate(
            img,
            level,
            interpolation=self.interpolation,
            expand=self.expand,
            center=self.center,
            fill=self.fill,
        )


class Shear(nn.Module):
    pixmix_max_level = 0.3
    level_type = float
    corruption_overlap = False

    def __init__(
        self,
        axis: int,
        random_direction: bool = True,
        interpolation: F.InterpolationMode = F.InterpolationMode.NEAREST,
        center: list[float] | None = None,
        fill: list[float] | None = None,
    ) -> None:
        super().__init__()
        if axis not in (0, 1):
            raise ValueError("Axis must be 0 or 1.")
        self.axis = axis
        self.random_direction = random_direction
        self.interpolation = interpolation
        self.center = center
        self.fill = fill

    def forward(self, img: Tensor | Image.Image, level: int) -> Tensor | Image.Image:
        if self.random_direction and torch.rand(1).item() > 0.5:  # coverage: ignore
            level = -level
        shear = [0.0, 0.0]
        shear[self.axis] = level
        return F.affine(
            img,
            angle=0,
            scale=1.0,
            shear=shear,
            translate=[0, 0],
            interpolation=self.interpolation,
            center=self.center,
            fill=self.fill,
        )


class Translate(nn.Module):
    pixmix_max_level = 0.45
    level_type = float
    corruption_overlap = False

    def __init__(
        self,
        axis: int,
        random_direction: bool = True,
        interpolation: F.InterpolationMode = F.InterpolationMode.NEAREST,
        center: list[float] | None = None,
        fill: list[float] | None = None,
    ) -> None:
        super().__init__()
        if axis not in (0, 1):
            raise ValueError("Axis must be 0 or 1.")
        self.axis = axis
        self.random_direction = random_direction
        self.interpolation = interpolation
        self.center = center
        self.fill = fill

    def forward(self, img: Tensor | Image.Image, level: float) -> Tensor | Image.Image:
        if self.random_direction and torch.rand(1).item() > 0.5:  # coverage: ignore
            level = -level
        translate = [0.0, 0.0]
        translate[self.axis] = level
        return F.affine(
            img,
            angle=0,
            scale=1.0,
            shear=[0, 0],
            translate=translate,
            interpolation=self.interpolation,
            center=self.center,
            fill=self.fill,
        )


class Contrast(nn.Module):
    pixmix_max_level = 1.8
    level_type = float
    corruption_overlap = True

    def __init__(self) -> None:
        super().__init__()

    def forward(self, img: Tensor | Image.Image, level: float) -> Tensor | Image.Image:
        if level < 0:
            raise ValueError("Level must be greater than 0.")
        return F.adjust_contrast(img, level)


class Brightness(nn.Module):
    pixmix_max_level = 1.8
    level_type = float
    corruption_overlap = True

    def __init__(self) -> None:
        super().__init__()

    def forward(self, img: Tensor, level: float) -> Tensor:
        if level < 0:
            raise ValueError("Level must be greater than 0.")
        return F.adjust_brightness(img, level)


class Saturation(nn.Module):
    level_type = float
    corruption_overlap = True

    def __init__(self) -> None:
        super().__init__()

    def forward(self, img: Tensor, level: float) -> Tensor:
        if level < 0:
            raise ValueError("Level must be greater than 0.")
        return F.adjust_saturation_image(img, level)


class Sharpen(nn.Module):
    pixmix_max_level = 1.8
    level_type = float
    corruption_overlap = True

    def __init__(self) -> None:
        super().__init__()

    def forward(self, img: Tensor | Image.Image, level: float) -> Tensor | Image.Image:
        if level < 0:
            raise ValueError("Level must be greater than 0.")
        return F.adjust_sharpness(img, level)


class Color(nn.Module):
    pixmix_max_level = 1.8
    level_type = float
    corruption_overlap = True

    def __init__(self) -> None:
        """Color augmentation class."""
        super().__init__()

    def forward(self, img: Tensor | Image.Image, level: float) -> Tensor | Image.Image:
        if level < 0:
            raise ValueError("Level must be greater than 0.")
        pil_img = F.to_pil_image(img) if isinstance(img, Tensor) else img
        pil_img = ImageEnhance.Color(pil_img).enhance(level)
        if isinstance(img, Tensor):
            return F.pil_to_tensor(pil_img)
        return pil_img


class RandomRescale(Transform):
    def __init__(
        self,
        min_scale: int,
        max_scale: int,
        interpolation: InterpolationMode | int = InterpolationMode.BILINEAR,
        antialias: bool | None = True,
    ) -> None:
        """Randomly rescale the input.

        This transformation can be used together with ``RandomCrop`` as data augmentations to train
        models on image segmentation task.

        Output spatial size is randomly sampled from the interval ``[min_size, max_size]``:

        .. code-block:: python

            scale = uniform_sample(min_scale, max_scale)
            output_width = input_width * scale
            output_height = input_height * scale

        If the input is a :class:`torch.Tensor` or a ``TVTensor`` (e.g. :class:`~torchvision.tv_tensors.Image`,
        :class:`~torchvision.tv_tensors.Video`, :class:`~torchvision.tv_tensors.BoundingBoxes` etc.)
        it can have arbitrary number of leading batch dimensions. For example,
        the image can have ``[..., C, H, W]`` shape. A bounding box can have ``[..., 4]`` shape.

        Args:
            min_scale (int): Minimum scale for random sampling
            max_scale (int): Maximum scale for random sampling
            interpolation (InterpolationMode, optional): Desired interpolation enum defined by
                :class:`torchvision.transforms.InterpolationMode`. Default is ``InterpolationMode.BILINEAR``.
                If input is Tensor, only ``InterpolationMode.NEAREST``, ``InterpolationMode.NEAREST_EXACT``,
                ``InterpolationMode.BILINEAR`` and ``InterpolationMode.BICUBIC`` are supported.
                The corresponding Pillow integer constants, e.g. ``PIL.Image.BILINEAR`` are accepted as well.
            antialias (bool, optional): Whether to apply antialiasing.
                It only affects **tensors** with bilinear or bicubic modes and it is
                ignored otherwise: on PIL images, antialiasing is always applied on
                bilinear or bicubic modes; on other modes (for PIL images and
                tensors), antialiasing makes no sense and this parameter is ignored.
                Possible values are:

                - ``True`` (default): will apply antialiasing for bilinear or bicubic modes.
                Other mode aren't affected. This is probably what you want to use.
                - ``False``: will not apply antialiasing for tensors on any mode. PIL
                images are still antialiased on bilinear or bicubic modes, because
                PIL doesn't support no antialias.
                - ``None``: equivalent to ``False`` for tensors and ``True`` for
                PIL images. This value exists for legacy reasons and you probably
                don't want to use it unless you really know what you are doing.

                The default value changed from ``None`` to ``True`` in
                v0.17, for the PIL and Tensor backends to be consistent.
        """
        super().__init__()
        self.min_scale = min_scale
        self.max_scale = max_scale
        self.interpolation = interpolation
        self.antialias = antialias

    # Compatibility with torchvision < 0.21
    def _get_params(self, flat_inputs: list[Any]) -> dict[str, Any]:
        height, width = query_size(flat_inputs)
        scale = torch.rand(1)
        scale = self.min_scale + scale * (self.max_scale - self.min_scale)
        return {"size": (int(height * scale), int(width * scale))}

    # Compatibility with torchvision >= 0.21
    def make_params(self, flat_inputs: list[Any]) -> dict[str, Any]:
        return self._get_params(flat_inputs)

    # Compatibility with torchvision < 0.21
    def _transform(self, inpt: Any, params: dict[str, Any]) -> Any:
        return self._call_kernel(
            F.resize,
            inpt,
            params["size"],
            interpolation=self.interpolation,
            antialias=self.antialias,
        )

    # Compatibility with torchvision >= 0.21
    def transform(self, inpt: Any, params: dict[str, Any]) -> Any:
        return self._transform(inpt, params)



from copy import deepcopy
from pathlib import Path

from PIL import Image
from torch import nn
from torchvision.datasets import VisionDataset
from torchvision.transforms import ToPILImage, ToTensor
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from torch_uncertainty.transforms.corruption import corruption_transforms


class CorruptedDataset(VisionDataset):
    def __init__(
        self,
        core_dataset: VisionDataset,
        shift_severity: int,
        on_the_fly: bool = False,
    ) -> None:
        super().__init__()
        self.core_dataset = core_dataset
        if shift_severity <= 0:
            raise ValueError(f"Severity must be greater than 0. Got {shift_severity}.")
        self.shift_severity = shift_severity
        self.core_length = len(core_dataset)
        self.on_the_fly = on_the_fly
        self.transforms = deepcopy(core_dataset.transforms)
        self.target_transforms = deepcopy(core_dataset.target_transform)
        self.core_dataset.transform = None
        self.core_dataset.target_transform = None

        self.root = Path(core_dataset.root)
        dataset_name = str(type(core_dataset)).split(".")[-1][:-2].lower()
        self.root /= dataset_name + "_corrupted"
        self.root /= f"severity_{self.shift_severity}"
        self.root.mkdir(parents=True)

        if not on_the_fly:
            self.to_tensor = ToTensor()
            self.to_pil = ToPILImage()
            self.samples = []
            self.targets = self.core_dataset.targets * 10
            self.prepare_data()

    def prepare_data(self):
        with logging_redirect_tqdm():
            for corruption in tqdm(corruption_transforms):
                corruption_name = corruption.__name__.lower()
                (self.root / corruption_name).mkdir(parents=True)
                self.save_corruption(self.root / corruption_name, corruption(self.shift_severity))

    def save_corruption(self, root: Path, corruption: nn.Module) -> None:
        for i in range(self.core_length):
            img, tgt = self.core_dataset[i]
            if isinstance(img, str | Path):
                img = Image.open(img).convert("RGB")
            img = corruption(self.to_tensor(img))
            self.to_pil(img).save(root / f"{i}.png")
            self.samples.append(root / f"{i}.png")
            self.targets.append(tgt)

    def __len__(self):
        """The length of the corrupted dataset."""
        return len(self.core_dataset) * len(corruption_transforms)

    def __getitem__(self, idx: int):
        """Get the corrupted image and the target.

        Args:
            idx (int): Index of the image to retrieve.
        """
        if self.on_the_fly:
            corrupt = corruption_transforms[idx // len(self.core_dataset)]
            idx = idx % len(self.core_dataset)
            img, target = self.core_dataset[idx]

            img = corrupt(img)
            if self.transform is not None:
                img = self.transform(img)

            if self.target_transform is not None:
                target = self.target_transform(target)
            return img, target

        img, target = self.core_dataset[idx]
        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target


if __name__ == "__main__":
    from torchvision.datasets import CIFAR10

    dataset = CIFAR10(root="data", download=True)
    corrupted_dataset = CorruptedDataset(dataset, shift_severity=1)
    print(len(corrupted_dataset))



# ruff: noqa: F401
from .checkpoints import get_version
from .cli import TULightningCLI
from .hub import load_hf
from .misc import create_train_val_split, csv_writer, plot_hist
from .trainer import TUTrainer
from .transforms import interpolation_modes_from_str



# ruff: noqa: F401
from .bayesian import ELBOLoss, KLDiv
from .classification import (
    BCEWithLogitsLSLoss,
    ConfidencePenaltyLoss,
    ConflictualLoss,
    DECLoss,
    FocalLoss,
)
from .regression import BetaNLL, DERLoss, DistributionNLLLoss



from numbers import Number

import torch
from torch import Tensor
from torch.distributions import (
    Cauchy,
    Distribution,
    Laplace,
    Normal,
    StudentT,
    constraints,
)
from torch.distributions.utils import broadcast_all


def get_dist_class(dist_family: str) -> type[Distribution]:
    """Get the distribution class from a string.

    Args:
        dist_family (str): The distribution family.

    Returns:
        type[Distribution]: The distribution class.
    """
    if dist_family == "normal":
        return Normal
    if dist_family == "laplace":
        return Laplace
    if dist_family == "nig":
        return NormalInverseGamma
    if dist_family == "cauchy":
        return Cauchy
    if dist_family == "student":
        return StudentT
    raise NotImplementedError(
        f"{dist_family} distribution is not supported. Raise an issue if needed."
    )


def get_dist_estimate(dist: Distribution, dist_estimate: str) -> Tensor:
    """Get a point-wise prediction from a distribution.

    Args:
        dist (Distribution): The distribution.
        dist_estimate (str): The estimate to use.

    Returns:
        Tensor: The estimated value.
    """
    if dist_estimate == "mean":
        return dist.mean
    if dist_estimate == "mode":
        return dist.mode
    raise NotImplementedError(
        f"{dist_estimate} estimate is not supported.Raise an issue if needed."
    )


class NormalInverseGamma(Distribution):
    arg_constraints = {
        "loc": constraints.real,
        "lmbda": constraints.positive,
        "alpha": constraints.greater_than_eq(1),
        "beta": constraints.positive,
    }
    support = constraints.real
    has_rsample = False

    def __init__(
        self,
        loc: Number | Tensor,
        lmbda: Number | Tensor,
        alpha: Number | Tensor,
        beta: Number | Tensor,
        validate_args: bool | None = None,
    ) -> None:
        self.loc, self.lmbda, self.alpha, self.beta = broadcast_all(loc, lmbda, alpha, beta)
        if (
            isinstance(loc, Number)
            and isinstance(lmbda, Number)
            and isinstance(alpha, Number)
            and isinstance(beta, Number)
        ):
            batch_shape = torch.Size()
        else:
            batch_shape = self.loc.size()
        super().__init__(batch_shape, validate_args=validate_args)

    @property
    def mean(self) -> Tensor:
        """Impromper mean of the NormalInverseGamma distribution.

        This value is necessary to perform point-wise predictions in the
        regression routine.
        """
        return self.loc

    def mode(self) -> None:
        raise NotImplementedError("NormalInverseGamma distribution has no mode.")

    def stddev(self) -> None:
        raise NotImplementedError("NormalInverseGamma distribution has no stddev.")

    def variance(self) -> None:
        raise NotImplementedError("NormalInverseGamma distribution has no variance.")

    @property
    def mean_loc(self) -> Tensor:
        return self.loc

    @property
    def mean_variance(self) -> Tensor:
        return self.beta / (self.alpha - 1)

    @property
    def variance_loc(self) -> Tensor:
        return self.beta / (self.alpha - 1) / self.lmbda

    def log_prob(self, value: Tensor) -> Tensor:
        if self._validate_args:  # coverage: ignore
            self._validate_sample(value)
        gam: Tensor = 2 * self.beta * (1 + self.lmbda)
        return (
            -0.5 * torch.log(torch.pi / self.lmbda)
            + self.alpha * gam.log()
            - (self.alpha + 0.5) * torch.log(gam + self.lmbda * (value - self.loc) ** 2)
            - torch.lgamma(self.alpha)
            + torch.lgamma(self.alpha + 0.5)
        )



# ruff: noqa: F401
from .aggregated_dataset import AggregatedDataset
from .fractals import Fractals
from .frost import FrostImages
from .kitti import KITTIDepth
from .muad import MUAD
from .nyu import NYUv2
